"""Functions related to reading and writing data."""

import io
import logging

# import argparse
from collections import Counter
from pathlib import Path

import requests

from . import cif, inputgen, pdb, psize
from . import definitions as defns
from .config import (
    AA_DEF_PATH,
    FILTER_WARNINGS,
    FILTER_WARNINGS_LIMIT,
    FORCE_FIELDS,
    NA_DEF_PATH,
    PATCH_DEF_PATH,
    TITLE_STR,
)
from .structures import Atom

_LOGGER = logging.getLogger(__name__)


class DuplicateFilter(logging.Filter):
    """Filter duplicate messages."""

    def __init__(self):
        super().__init__()
        self.warn_count = Counter()

    def filter(self, record):
        """Filter current record."""
        if record.levelname == "WARNING":
            for fwarn in FILTER_WARNINGS:
                if record.getMessage().startswith(fwarn):
                    self.warn_count.update([fwarn])
                    if self.warn_count[fwarn] > FILTER_WARNINGS_LIMIT:
                        return False
                    elif self.warn_count[fwarn] == FILTER_WARNINGS_LIMIT:
                        _LOGGER.warning(
                            f'Suppressing further "{fwarn}" messages'
                        )
                        return False
                    else:
                        return True
        return True


def print_biomolecule_atoms(atomlist, chainflag=False, pdbfile=False):
    """Get PDB-format text lines for specified atoms.

    :param [Atom] atomlist:  the list of atoms to include
    :param bool chainflag:  flag whether to print chainid or not
    :return:  list of strings, each representing an atom PDB line
    :rtype:  [str]
    """
    text = []
    currentchain_id = None
    for iatom, atom in enumerate(atomlist):
        # Print the "TER" records between chains
        if currentchain_id is None:
            currentchain_id = atom.chain_id
        elif atom.chain_id != currentchain_id:
            currentchain_id = atom.chain_id
            text.append("TER\n")
        atom.serial = iatom + 1
        if pdbfile is True:
            text.append(f"{atom.get_pdb_string()}\n")
        else:
            text.append(f"{atom.get_pqr_string(chainflag=chainflag)}\n")
    text.append("TER\nEND")
    return text


def get_old_header(pdblist):
    """Get old header from list of :mod:`pdb` objects.

    :param pdblist:  list of :mod:`pdb` block objects
    :type pdblist:  []
    :return:  old header as string
    :rtype:  str
    """
    old_header = io.StringIO()
    header_types = (
        pdb.HEADER,
        pdb.TITLE,
        pdb.COMPND,
        pdb.SOURCE,
        pdb.KEYWDS,
        pdb.EXPDTA,
        pdb.AUTHOR,
        pdb.REVDAT,
        pdb.JRNL,
        pdb.REMARK,
        pdb.SPRSDE,
        pdb.NUMMDL,
    )
    for pdb_obj in pdblist:
        if not isinstance(pdb_obj, header_types):
            break
        old_header.write(str(pdb_obj))
        old_header.write("\n")
    return old_header.getvalue()


def print_pqr_header(
    pdblist,
    atomlist,
    reslist,
    charge,
    force_field,
    ph_calc_method,
    ph,
    ffout,
    include_old_header=False,
):
    """Print the header for the PQR file.

    :param pdblist:  list of lines from original PDB with header
    :type pdblist:  [str]
    :param atomlist:  a list of atoms that were unable to have charges assigned
    :type atomlist:  [Atom]
    :param reslist:  a list of residues with non-integral charges
    :type reslist:  [Residue]
    :param charge:  the total charge on the biomolecule
    :type charge:  float
    :param force_field:  the forcefield name
    :type force_field:  str
    :param ph_calc_method:  pKa calculation method
    :type ph_calc_method:  str
    :param ph:  pH value, if any
    :type ph:  float
    :param ffout:  forcefield used for naming scheme
    :type ffout:  str
    :return:  the header for the PQR file
    :rtype:  str
    """
    if force_field is None:
        force_field = "User force field"
    else:
        force_field = force_field.upper()
    head = "\n".join(
        [
            "REMARK   1 PQR file generated by PDB2PQR",
            f"REMARK   1 {TITLE_STR}",
            "REMARK   1",
            f"REMARK   1 Forcefield Used: {force_field}",
            "",
        ]
    )
    if ffout is not None:
        head += f"REMARK   1 Naming Scheme Used: {ffout}\n"
    head += head + "REMARK   1\n"
    if ph_calc_method is not None:
        head += "\n".join(
            [
                f"REMARK   1 pKas calculated by {ph_calc_method} and "
                f"assigned using pH {ph:.2f}\n"
                "REMARK   1",
                "",
            ]
        )
    if len(atomlist) != 0:
        head += "\n".join(
            [
                "REMARK   5 WARNING: PDB2PQR was unable to assign charges",
                "REMARK   5 to the following atoms (omitted below):",
                "",
            ]
        )
        for atom in atomlist:
            head += f"REMARK   5    {atom.serial} {atom.name} in "
            head += f"{atom.residue.name} {atom.residue.res_seq}\n"
        head += "\n".join(
            [
                "REMARK   5 This is usually due to the fact that this residue "
                "is not",
                "REMARK   5 an amino acid or nucleic acid; or, there are no "
                "parameters",
                "REMARK   5 available for the specific protonation state of "
                "this",
                "REMARK   5 residue in the selected forcefield.",
                "REMARK   5",
                "",
            ]
        )
    if len(reslist) != 0:
        head += "\n".join(
            [
                "REMARK   5 WARNING: Non-integral net charges were found in",
                "REMARK   5 the following residues:",
                "",
            ]
        )
        for residue in reslist:
            head += f"REMARK   5    {residue} - "
            head += f"Residue Charge: {residue.charge:.4f}\n"
        head += "REMARK   5\n"
    head += "\n".join(
        [
            f"REMARK   6 Total charge on this biomolecule: {charge:.4f} e\n"
            "REMARK   6",
            "",
        ]
    )
    if include_old_header:
        head += "\n".join(
            ["REMARK   7 Original PDB header follows", "REMARK   7", ""]
        )
        head += get_old_header(pdblist)
    return head


def print_pqr_header_cif(
    atomlist,
    reslist,
    charge,
    force_field,
    ph_calc_method,
    ph,
    ffout,
    include_old_header=False,
):
    """Print the header for the PQR file in CIF format.

    :param atomlist:  a list of atoms that were unable to have charges assigned
    :type atomlist:  [Atom]
    :param reslist:  a list of residues with non-integral charges
    :type reslist:  [Residue]
    :param charge:  the total charge on the biomolecule
    :type charge:  float
    :param force_field:  the forcefield name
    :type force_field:  str
    :param ph_calc_method:  pKa calculation method
    :type ph_calc_method:  str
    :param ph:  pH value, if any
    :type ph:  float
    :param ffout:  forcefield used for naming scheme
    :type ffout:  str
    :return:  the header for the PQR file
    :rtype:  str
    """
    if force_field is None:
        force_field = "User force field"
    else:
        force_field = force_field.upper()

    header = "\n".join(
        [
            "#",
            "loop_",
            "_pdbx_database_remark.id",
            "_pdbx_database_remark.text",
            "1",
            ";",
            "PQR file generated by PDB2PQR",
            TITLE_STR,
            "",
            f"Forcefield used: {force_field}\n",
            "",
        ]
    )
    if ffout is not None:
        header += f"Naming scheme used: {ffout}\n"
    header += "\n"
    if ph_calc_method is not None:
        header += f"pKas calculated by {ph_calc_method} and "
        header += f"assigned using pH {ph:.2f}\n"
    header += "\n".join([";", "2", ";", ""])
    if len(atomlist) > 0:
        header += "\n".join(
            [
                "Warning: PDB2PQR was unable to assign charges",
                "to the following atoms (omitted below):",
                "",
            ]
        )
        for atom in atomlist:
            header += f"    {atom.serial} {atom.name} in "
            header += f"{atom.residue.name} {atom.residue.res_seq}\n"
        header += "\n".join(
            [
                "This is usually due to the fact that this residue is not",
                "an amino acid or nucleic acid; or, there are no parameters",
                "available for the specific protonation state of this",
                "residue in the selected forcefield.",
                "",
            ]
        )
    if len(reslist) > 0:
        header += "\n".join(
            [
                "Warning: Non-integral net charges were found in",
                "the following residues:",
                "",
            ]
        )
        for residue in reslist:
            header += f"    {residue} - "
            header += f"Residue Charge: {residue.charge:.4f}\n"
    header += "\n".join(
        [
            ";",
            "3",
            ";",
            f"Total charge on this biomolecule: {charge:.4f} e",
            ";",
            "",
        ]
    )
    if include_old_header:
        _LOGGER.warning("Including original CIF header not implemented.")
    header += "\n".join(
        [
            "#",
            "loop_",
            "_atom_site.group_PDB",
            "_atom_site.id",
            "_atom_site.label_atom_id",
            "_atom_site.label_comp_id",
            "_atom_site.label_seq_id",
            "_atom_site.Cartn_x",
            "_atom_site.Cartn_y",
            "_atom_site.Cartn_z",
            "_atom_site.pqr_partial_charge",
            "_atom_site.pqr_radius",
            "",
        ]
    )
    return header


def dump_apbs(output_pqr, output_path):
    """Generate and dump APBS input files related to output_pqr.

    :param output_pqr:  path to PQR file used to generate APBS input file
    :type output_pqr:  str
    :param output_path:  path for APBS input file output
    :type output_path:  str
    """
    method = "mg-auto"
    size = psize.Psize()
    size.parse_input(output_pqr)
    size.run_psize(output_pqr)
    input_ = inputgen.Input(output_pqr, size, method, 0, potdx=True)
    input_.print_input_files(output_path)


def test_for_file(name, type_):
    """Test for the existence of a file with a few name permutations.

    :param name:  name of file
    :type name:  str
    :param type_:  type of file
    :type type_:  str
    :return:  path to file
    :raises FileNotFoundError:  if file not found
    :rtype:  Path
    """
    if name is None:
        return ""
    test_names = [name, name.upper(), name.lower()]
    test_suffixes = ["", f".{type_.upper()}", f".{type_.lower()}"]

    module_path = Path(__file__)
    dirpath_dat = module_path.parent / "dat"

    assert dirpath_dat.is_dir()

    if name.lower() in FORCE_FIELDS:
        name = name.upper()
    for test_name in test_names:
        for test_suffix in test_suffixes:
            test_path = dirpath_dat / (test_name + test_suffix)
            if test_path.is_file():
                _LOGGER.debug(f"Found {type_} file {test_path}")
                return test_path
    err = f"Unable to find {type_} file for {name}"
    raise FileNotFoundError(err)


def test_names_file(name):
    """Test for the .names file that contains the XML mapping.

    :param name:  the name of the forcefield
    :type name:  str
    :returns:  the path to the file
    :rtype:  Path
    :raises FileNotFoundError:  file not found
    """
    return test_for_file(name, "NAMES")


def test_dat_file(name):
    """Test for the forcefield file with a few name permutations.

    :param name:  the name of the dat file
    :type name:  str
    :returns:  the path to the file
    :rtype:  Path
    :raises FileNotFoundError:  file not found
    """
    return test_for_file(name, "DAT")


def test_xml_file(name):
    """Test for the XML file with a few name permutations.

    :param name:  the name of the dat file
    :type name:  str
    :returns:  the path to the file
    :rtype:  Path
    :raises FileNotFoundError:  file not found
    """
    return test_for_file(name, "xml")


def get_pdb_file(name):
    """Obtain a PDB file.

    First check the path given on the command line - if that file is not
    available, obtain the file from the PDB webserver at
    http://www.rcsb.org/pdb/

    .. todo::  This should be a context manager (to close the open file).
    .. todo::  Remove hard-coded parameters.

    :param name:  name of PDB file (path) or PDB ID
    :type name:  str
    :return:  file-like object containing PDB file
    :rtype:  file
    """
    path = Path(name)
    if path.is_file():
        return open(path, encoding="utf-8")
    url_path = f"https://files.rcsb.org/download/{path.stem}.pdb"
    _LOGGER.debug(f"Attempting to fetch PDB from {url_path}")
    resp = requests.get(url_path)
    if resp.status_code != requests.codes["ok"]:
        errstr = f"Got code {resp.status_code} while retrieving {url_path}"
        raise IOError(errstr)
    return io.StringIO(resp.text)


def get_molecule(input_path):
    """Get molecular structure information as a series of parsed lines.

    :param input_path:  structure file PDB ID or path
    :type intput_path:  str
    :return: (list of molecule records, Boolean indicating whether entry is
        CIF)
    :rtype:  ([str], bool)
    :raises RuntimeError:  problems with structure file
    """
    path = Path(input_path)
    input_file = get_pdb_file(input_path)
    is_cif = False
    if path.suffix.lower() == ".cif":
        pdblist, errlist = cif.read_cif(input_file)
        is_cif = True
    else:
        pdblist, errlist = pdb.read_pdb(input_file)
    input_file.close()
    if len(pdblist) == 0 and len(errlist) == 0:
        raise RuntimeError(f"Unable to find file {path}!")
    if len(errlist) != 0:
        if is_cif:
            _LOGGER.warning(f"Warning: {path} is a non-standard CIF file.\n")
        else:
            _LOGGER.warning(f"Warning: {path} is a non-standard PDB file.\n")
        _LOGGER.error(errlist)
    return pdblist, is_cif


def get_definitions(
    aa_path=AA_DEF_PATH, na_path=NA_DEF_PATH, patch_path=PATCH_DEF_PATH
):
    """Load topology definition files.

    :param aa_path:  likely location of amino acid topology definitions
    :type aa_path:  str
    :param na_path:  likely location of nucleic acid topology definitions
    :type na_path:  str
    :param patch_path:  likely location of patch topology definitions
    :type patch_path:  str
    :return:  topology Definitions object.
    :rtype:  Definition
    """
    aa_path_ = test_xml_file(aa_path)
    na_path_ = test_xml_file(na_path)
    patch_path_ = test_xml_file(patch_path)
    with (
        open(aa_path_) as aa_file,
        open(na_path_) as na_file,
        open(patch_path_) as patch_file,
    ):
        definitions = defns.Definition(
            aa_file=aa_file, na_file=na_file, patch_file=patch_file
        )
    return definitions


def setup_logger(output_pqr, level="DEBUG"):
    """Setup the logger.

    Setup logger to output the log file to the same directory as PQR
    output.

    :param str output_pqr:  path to PQR file
    :param str level:  logging level
    """
    # Get the output logging location
    output_pth = Path(output_pqr)
    log_file = Path(output_pth.parent, output_pth.stem + ".log")
    log_format = (
        "%(asctime)s %(levelname)s:%(filename)s:"
        "%(lineno)d:%(funcName)s:%(message)s"
    )
    _LOGGER.info(f"Logs stored: {log_file}")
    logging.basicConfig(
        filename=log_file,
        format=log_format,
        level=getattr(logging, level),
    )
    console = logging.StreamHandler()
    formatter = logging.Formatter("%(levelname)s:%(message)s")
    console.setFormatter(formatter)
    console.setLevel(level=getattr(logging, level))
    logging.getLogger("").addHandler(console)
    logging.getLogger("").addFilter(DuplicateFilter())


def read_pqr(pqr_file):
    """Read PQR file.

    :param pqr_file:  file object ready for reading as text
    :type pqr_file:  file
    :returns:  list of atoms read from file
    :rtype:  [Atom]
    """
    atoms = []
    for line in pqr_file:
        atom = Atom.from_pqr_line(line)
        if atom is not None:
            atoms.append(atom)
    return atoms


def read_qcd(qcd_file):
    """Read QCD (UHDB QCARD format) file.

    :param file qcd_file:  file object ready for reading as text
    :returns:  list of atoms read from file
    :rtype:  [Atom]
    """
    atoms = []
    atom_serial = 1
    for line in qcd_file:
        atom = Atom.from_qcd_line(line, atom_serial)
        if atom is not None:
            atoms.append(atom)
            atom_serial += 1
    return atoms


def read_dx(dx_file):
    """Read DX-format volumetric information.

    The OpenDX file format is defined at
    <https://www.idvbook.com/wp-content/uploads/2010/12/opendx.pdf`.

    .. note:: This function is not a general-format OpenDX file parser and
       makes many assumptions about the input data type, grid structure, etc.

    .. todo:: This function should be moved into the APBS code base.

    :param dx_file:  file object for DX file, ready for reading as text
    :type dx_file:  file
    :returns:  dictionary with data from DX file
    :rtype:  dict
    :raises ValueError:  on parsing error
    """
    dx_dict = {
        "grid spacing": [],
        "values": [],
        "number of grid points": None,
        "lower left corner": None,
    }
    for line in dx_file:
        words = [w.strip() for w in line.split()]
        if words[0] in ["#", "attribute", "component"]:
            pass
        elif words[0] == "object":
            if words[1] == "1":
                dx_dict["number of grid points"] = (
                    int(words[5]),
                    int(words[6]),
                    int(words[7]),
                )
        elif words[0] == "origin":
            dx_dict["lower left corner"] = [
                float(words[1]),
                float(words[2]),
                float(words[3]),
            ]
        elif words[0] == "delta":
            spacing = [float(words[1]), float(words[2]), float(words[3])]
            dx_dict["grid spacing"].append(spacing)
        else:
            for word in words:
                dx_dict["values"].append(float(word))
    return dx_dict


def write_cube(cube_file, data_dict, atom_list, comment="CPMD CUBE FILE."):
    """Write a Cube-format data file.

    Cube file format is defined at
    <https://docs.chemaxon.com/display/Gaussian_Cube_format.html>.

    .. todo:: This function should be moved into the APBS code base.

    :param cube_file:  file object ready for writing text data
    :type cube_file:  file
    :param data_dict:  dictionary of volumetric data as produced by
        :func:`read_dx`
    :type data_dict:  dict
    :param comment:  comment for Cube file
    :type comment:  str
    """
    cube_file.write(comment + "\n")
    cube_file.write("OUTER LOOP: X, MIDDLE LOOP: Y, INNER LOOP: Z\n")
    num_atoms = len(atom_list)
    origin = data_dict["lower left corner"]
    cube_file.write(
        f"{num_atoms:>4} {origin[0]:>11.6f} {origin[1]:>11.6f} "
        f"{origin[2]:>11.6f}\n"
    )
    num_points = data_dict["number of grid points"]
    spacings = data_dict["grid spacing"]
    for i in range(3):
        cube_file.write(
            f"{-num_points[i]:>4} "
            f"{spacings[i][0]:>11.6f} "
            f"{spacings[i][1]:>11.6f} "
            f"{spacings[i][2]:>11.6f}\n"
        )
    for atom in atom_list:
        cube_file.write(
            f"{atom.serial:>4} {atom.charge:>11.6f} {atom.x:>11.6f} "
            f"{atom.y:>11.6f} {atom.z:>11.6f}\n"
        )
    stride = 6
    values = data_dict["values"]
    for i in range(0, len(values), 6):
        if i + stride < len(values):
            imax = i + 6
            words = [f"{val:< 13.5E}" for val in values[i:imax]]
            cube_file.write(" ".join(words) + "\n")
        else:
            words = [f"{val:< 13.5E}" for val in values[i:]]
            cube_file.write(" ".join(words))



"""Regression tests for PDB2PQR behavior."""

import logging

import common
import pytest

_LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "args, input_pdb, output_pqr, expected_pqr",
    [
        pytest.param(
            "--log-level=INFO --ff=AMBER",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_ff=AMBER.pqr",
            id="1AFS basic local",
        ),
        pytest.param(
            "--log-level=INFO --ff=AMBER",
            "1AFS",
            "output.pqr",
            common.DATA_DIR / "1AFS_ff=AMBER.pqr",
            id="1AFS basic remote",
        ),
    ],
)
def test_basic(args, input_pdb, output_pqr, expected_pqr, tmp_path):
    """Basic code to run 1AFS."""
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        expected_pqr=expected_pqr,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize(
    "args, input_pdb, output_pqr, expected_pqr",
    [
        pytest.param(
            "--log-level=INFO --whitespace --ff=AMBER",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_whitespace_ff=AMBER.pqr",
            id="1AFS whitespace AMBER",
        ),
        pytest.param(
            "--log-level=INFO --whitespace --ff=CHARMM",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_whitespace_ff=CHARMM.pqr",
            id="1AFS whitespace CHARMM",
        ),
        pytest.param(
            "--log-level=INFO --whitespace --ff=PARSE",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_whitespace_ff=PARSE.pqr",
            id="1AFS whitespace PARSE",
        ),
        pytest.param(
            "--log-level=INFO --whitespace --ff=PEOEPB",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_whitespace_ff=PEOEPB.pqr",
            id="1AFS whitespace PEOEPB",
        ),
        pytest.param(
            "--log-level=INFO --whitespace --ff=SWANSON",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_whitespace_ff=SWANSON.pqr",
            id="1AFS whitespace SWANSON",
        ),
        pytest.param(
            "--log-level=INFO --whitespace --ff=TYL06",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_whitespace_ff=TYL06.pqr",
            id="1AFS whitespace TYL06",
        ),
    ],
)
def test_forcefields(args, input_pdb, output_pqr, expected_pqr, tmp_path):
    """Basic code to run 1AFS with --whitespace for different forcefields."""
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        expected_pqr=expected_pqr,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize(
    "args, input_pdb, output_pqr, expected_pqr",
    [
        pytest.param(
            "--log-level=INFO --whitespace --clean",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_clean_whitespace.pqr",
            id="1AFS whitespace clean",
        ),
        pytest.param(
            "--log-level=INFO --whitespace --assign-only --ff=AMBER",
            common.DATA_DIR / "1A1P.pdb",
            "output.pqr",
            common.DATA_DIR / "1A1P_assign-only_whitespace_ff=AMBER.pqr",
            id="1A1P assign-only whitespace AMBER",
        ),
        pytest.param(
            "--log-level=INFO --whitespace --nodebump --noopt --ff=AMBER",
            common.DATA_DIR / "1A1P.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_nodebump_noopt_whitespace_ff=AMBER.pqr",
            id="1A1P nodebump noopt whitespace AMBER",
        ),
        pytest.param(
            "--ff=AMBER --keep-chain --whitespace",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_chain_whitespace_ff=AMBER.pqr",
            id="1AFS chain whitespace AMBER",
        ),
        pytest.param(
            "--ff=PARSE --neutraln --neutralc --whitespace",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_neutralc_neutraln_whitespace_ff=PARSE.pqr",
            id="1AFS neutralc neutraln PARSE",
        ),
        pytest.param(
            "--ff=AMBER --drop-water",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_drop-water_ff=AMBER.pqr",
            id="1AFS drop-water AMBER",
        ),
        pytest.param(
            f"--userff={common.DATA_DIR / 'custom-ff.dat'} "
            f"--usernames={common.DATA_DIR / 'custom.names'} --whitespace",
            common.DATA_DIR / "1AFS.pdb",
            "output.pqr",
            common.DATA_DIR / "1AFS_userff_usernames_whitespace.pqr",
            id="1AFS userff usernames whitespace",
        ),
    ],
)
def test_other_options(args, input_pdb, output_pqr, expected_pqr, tmp_path):
    """Basic code to run 1AFS with --whitespace."""
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        expected_pqr=expected_pqr,
        tmp_path=tmp_path,
    )


# @pytest.mark.parametrize(
#   "args, input_pdb, output_pqr, expected_pqr",
#   pytest.param(
#   "--ff=AMBER --apbs-input --whitespace --include-header",
#   common.DATA_DIR/"1AFS.pdb",
#   "output.pqr",
#   common.DATA_DIR/'1AFS_apbs-input_include-header_whitespace_ff=AMBER.pqr',
#     id="1AFS apbs-input include-header whitespace AMBER"
#   ),
_LOGGER.error(
    "Need to reinstate the --apbs-input test with a temporary directory "
    "as path"
)



"""Common routines and variables for testing"""

import hashlib
import itertools
import logging
from pathlib import Path

import numpy as np
import pandas as pd

from pdb2pqr.main import build_main_parser, main_driver
from pdb2pqr.structures import Atom

_LOGGER = logging.getLogger(__name__)
PARSER = build_main_parser()


# This is a list of legit FF names
FF_LIST = ["AMBER", "CHARMM", "PARSE", "TYL06", "PEOEPB", "SWANSON"]
# Where the test data lives
DATA_DIR = Path("tests/data")
# Tolerable error for positions
POS_CUT = 1e-2
# Tolerable error for charge
Q_CUT = 1e-2
# Tolerable error for radius
R_CUT = 1e-2


def generate_ff_combinations():
    """Generate combinations of force field options.

    Returns:
        list of combinations
    """
    ff_options = [None] + [f"--ff={ff}" for ff in FF_LIST]
    ffout_options = [None] + [f"--ffout={ff}" for ff in FF_LIST]
    ff_combos = []
    for ff_opt in ff_options:
        for ffout_opt in ffout_options:
            combo = []
            if ff_opt is not None:
                combo += [ff_opt]
            if ffout_opt is not None:
                combo += [ffout_opt]
            ff_combos += [combo]
    return ff_combos


def generate_combinations(option_list):
    """Generate all combinations of provided options.

    Returns:
        List of option combinations.
    """
    combos = []
    for num in range(len(option_list) + 1):
        combos += list(itertools.combinations(option_list, num))
    return combos


def pqr_to_dict(pqr_file):
    """Convert PQR to dictionary.

    Args:
        pqr_file:  file object for PQR.

    Returns:
        DataFrame with PQR information.
    """
    pqr = []
    for line in pqr_file:
        atom = Atom.from_pqr_line(line)
        if atom is not None:
            row_dict = {"atom_num": atom.serial}
            # Many hydrogens are created in arbitrary order when attached to
            # the same heavy atom. Therefore, the last number in their name is
            # not meaningful
            atom_name = atom.name
            if atom_name[0] == "H":
                try:
                    int(atom_name[-1])
                    atom_name = atom_name[:-1] + "_"
                except ValueError:
                    pass
            row_dict["atom_name"] = atom_name
            row_dict["res_name"] = atom.res_name
            row_dict["res_num"] = atom.res_seq
            row_dict["x"] = atom.x
            row_dict["y"] = atom.y
            row_dict["z"] = atom.z
            row_dict["q"] = atom.charge
            row_dict["r"] = atom.radius
            pqr.append(row_dict)
    return pd.DataFrame(pqr)


def pqr_distance(df1, df2):
    """Calculate distances between positions, charges, and radii from two PQR dataframes.

    Args:
        df1:  PQR dataframe
        df2:  PQR dataframe

    Returns:
        Dataframe of distances
    """
    if "chain" in df1.columns:
        d_frame = df1.merge(
            df2,
            on=["atom_name", "res_name", "res_num", "chain"],
            how="inner",
            suffixes=("A", "B"),
        )
    else:
        d_frame = df1.merge(
            df2,
            on=["atom_name", "res_name", "res_num"],
            how="inner",
            suffixes=("A", "B"),
        )

    # Calculate differences and drop original columns
    for c_val in ("x", "y", "z", "q", "r"):
        d_val = f"d{c_val}"
        d2_val = f"d{c_val}2"
        c_a = f"{c_val}A"
        c_b = f"{c_val}B"
        d_frame[d_val] = d_frame[c_a] - d_frame[c_b]
        d_frame[d2_val] = d_frame[d_val] * d_frame[d_val]
        d_frame = d_frame.drop([d_val, c_a, c_b], axis="columns")

    # Calculate position norm-squared and drop used columns
    d_frame["dp2"] = d_frame["dx2"] + d_frame["dy2"] + d_frame["dz2"]
    d_frame = d_frame.drop(["dx2", "dy2", "dz2"], axis="columns")

    # Calculate norms of all measures and drop used columns
    for c_val in ("p", "q", "r"):
        n_val = f"d{c_val}"
        n2_val = f"d{c_val}2"
        d_frame[n_val] = np.sqrt(d_frame[n2_val])
        d_frame = d_frame.drop(n2_val, axis="columns")

    return d_frame


def compare_pqr(pqr1_path, pqr2_path, compare_resnames=False):
    """Compare two PQR files.

    Assume that atom numbering/ordering does not matter.

    Args:
        pqr1_path:  Path to first PQR
        par2_path:  Path to second PQR
    """
    with open(pqr1_path, encoding="utf-8") as pqr1_file:
        df1 = pqr_to_dict(pqr1_file)
        _LOGGER.debug(f"PQR 1 has shape {df1.shape}")

    with open(pqr2_path, encoding="utf-8") as pqr2_file:
        df2 = pqr_to_dict(pqr2_file)
        _LOGGER.debug(f"PQR 2 has shape {df2.shape}")

    if compare_resnames:
        diff_resname = df1.res_name != df2.res_name
        if any(diff_resname):
            _LOGGER.error(f"Got residue names:\n{df1[diff_resname]}")
            _LOGGER.error(f"Expected residue names:\n{df2[diff_resname]}")
            raise ValueError(
                f"PQR files have different residue names\n{df1.res_name[diff_resname]}\nreference\n{df2.res_name[diff_resname]}"
            )

    d_frame = pqr_distance(df1, df2)
    _LOGGER.debug(f"Merged d_frame has shape {d_frame.shape}")

    grouped = d_frame.groupby(["res_name", "res_name", "atom_name"])
    _LOGGER.debug(f"Have {len(grouped):d} unique atoms")
    df_min = grouped.min()

    for col, what, cut in [
        ("dp", "position", POS_CUT),
        ("dq", "charge", Q_CUT),
        ("dr", "radius", R_CUT),
    ]:
        for cut_ in [0.0, cut]:
            df_c = df_min[df_min[col] > cut_].sort_values(col, ascending=False)
            ndiff = df_c.shape[0]
            result = f"{ndiff:d} atoms have {what} differences > {cut_:g}"
            if ndiff > 0:
                _LOGGER.warning(result)
                df_c = df_min[df_min[col] > cut_].sort_values(
                    col, ascending=False
                )
                summary = [
                    f"{key}: {val:.3E}"
                    for (key, val) in df_c[col].describe().to_dict().items()
                ]
                _LOGGER.debug(summary)
                if cut_ > 0:
                    raise ValueError(result)
            else:
                _LOGGER.info(result)


def run_pdb2pqr_for_tests(
    args,
    input_pdb,
    tmp_path,
    output_pqr=None,
    expected_pqr=None,
    compare_resnames=False,
):
    """Basic code for invoking PDB2PQR."""
    from pdb2pqr import io

    if output_pqr is None:
        hash_str = f"{args}{input_pdb}"
        hash_ = hashlib.sha1(hash_str.encode("UTF-8")).hexdigest()
        output_pqr = f"{hash_}.pqr"
    output_pqr = tmp_path / output_pqr
    _LOGGER.debug(f"Writing output to {output_pqr}")
    arg_str = f"{args} {input_pdb} {output_pqr}"
    args = PARSER.parse_args(arg_str.split())
    io.setup_logger(args.output_pqr, args.log_level)
    _ = main_driver(args)
    if expected_pqr is not None:
        compare_pqr(
            output_pqr, expected_pqr, compare_resnames=compare_resnames
        )


def run_propka_for_tests(input_pdb, compare_file, pH):
    from propka.lib import build_parser

    from pdb2pqr import io
    from pdb2pqr.main import run_propka, setup_molecule

    propka_parser = build_parser()
    propka_parser.add_argument(
        "--keep-chain", action="store_true", default=False
    )
    propka_args = propka_parser.parse_args(
        ["--log-level", "WARNING", "--pH", str(pH), "TEST"]
    )

    definition = io.get_definitions()
    pdblist, _ = io.get_molecule(input_pdb)
    biomolecule, definition, _ = setup_molecule(pdblist, definition, None)
    biomolecule.set_termini(neutraln=False, neutralc=False)
    biomolecule.update_bonds()

    biomolecule.remove_hydrogens()
    pka_df, _ = run_propka(propka_args, biomolecule)
    results = {
        f"{row['res_name']} {row['res_num']} {row['chain_id']}": row["pKa"]
        for row in pka_df
        if row["group_label"].startswith(row["res_name"])
    }

    # # Used for writing the test CSV files
    # with open(compare_file, "w") as f:
    #     f.write("group,pka\n")
    #     for key, val in results.items():
    #         f.write(f"{key},{val}\n")

    compare = {}
    with open(compare_file) as f:
        for line in f.readlines()[1:]:
            line = line.strip()
            if len(line) == 0:
                continue
            key, val = line.split(",")
            compare[key] = float(val)

    diff_new = [key for key in results if key not in compare]
    if len(diff_new):
        raise RuntimeError(
            f"New pKa groups {' '.join(diff_new)} were generated by test which didn't exist in the test reference files."
        )
    diff_old = [key for key in compare if key not in results]
    if len(diff_old):
        raise RuntimeError(
            f"Missing pKa groups {' '.join(diff_old)} which exist in the test reference files."
        )

    for key in results:
        if abs(results[key] - compare[key]) > 1e-3:
            raise RuntimeError(
                f"PROPKA test error. pKa values changed more than 1e-3 for group {key}: new {results[key]} reference {compare[key]}"
            )


def run_pkaani_for_tests(input_pdb, compare_file, pH):
    from propka.lib import build_parser

    from pdb2pqr import io
    from pdb2pqr.main import run_pkaani, setup_molecule

    propka_parser = build_parser()
    propka_parser.add_argument(
        "--keep-chain", action="store_true", default=False
    )
    pka_args = propka_parser.parse_args(
        ["--log-level", "WARNING", "--pH", str(pH), "TEST"]
    )

    definition = io.get_definitions()
    pdblist, _ = io.get_molecule(input_pdb)
    biomolecule, definition, _ = setup_molecule(pdblist, definition, None)
    biomolecule.set_termini(neutraln=False, neutralc=False)
    biomolecule.update_bonds()

    biomolecule.remove_hydrogens()
    pka_df = run_pkaani(pka_args, biomolecule)
    results = {
        f"{row['res_name']} {row['res_num']} {row['chain_id']}": row["pKa"]
        for row in pka_df
    }

    # Used for writing the test CSV files
    with open(compare_file, "w") as f:
        f.write("group,pka\n")
        for key, val in results.items():
            f.write(f"{key},{val}\n")

    compare = {}
    with open(compare_file) as f:
        for line in f.readlines()[1:]:
            line = line.strip()
            if len(line) == 0:
                continue
            key, val = line.split(",")
            compare[key] = float(val)

    diff_new = [key for key in results if key not in compare]
    if len(diff_new):
        raise RuntimeError(
            f"New pKa groups {' '.join(diff_new)} were generated by test which didn't exist in the test reference files."
        )
    diff_old = [key for key in compare if key not in results]
    if len(diff_old):
        raise RuntimeError(
            f"Missing pKa groups {' '.join(diff_old)} which exist in the test reference files."
        )

    for key in results:
        if abs(results[key] - compare[key]) > 1e-3:
            raise RuntimeError(
                f"pKa-ANI test error. pKa values changed more than 1e-3 for group {key}: new {results[key]} reference {compare[key]}"
            )



"""Tests for PKAANI functionality."""

from pathlib import Path

import common
import pytest

PKAANI_TEST_DIR = Path("tests/pkaani_tests")


@pytest.mark.parametrize(
    "input_pdb",
    [
        str(PKAANI_TEST_DIR / "1brs_clean.pdb"),
        str(PKAANI_TEST_DIR / "1zy8_clean.pdb"),
        str(PKAANI_TEST_DIR / "6oge_de.pdb"),
        "1FCC.pdb",
    ],
    ids=str,
)
def test_pkaani_apo(input_pdb, tmp_path):
    """PKAANI non-regression tests on biomolecules without ligands."""
    args = (
        "--log-level=INFO --ff=AMBER --drop-water "
        "--titration-state-method=pkaani"
    )
    output_pqr = Path(input_pdb).stem + ".pqr"
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize(
    "input_pdb",
    [
        str(PKAANI_TEST_DIR / "1brs_clean.pdb"),
        str(PKAANI_TEST_DIR / "1zy8_clean.pdb"),
        str(PKAANI_TEST_DIR / "6oge_de.pdb"),
        "1FCC.pdb",
    ],
    ids=str,
)
def test_pkaani_pka(input_pdb):
    """PKAANI non-regression tests for pKa values on biomolecules without ligands."""
    pdb_filename = Path(input_pdb).name
    output_csv = PKAANI_TEST_DIR / f"{pdb_filename}_pka.csv"
    common.run_pkaani_for_tests(
        input_pdb=input_pdb, compare_file=output_csv, pH=7.4
    )



"""Tests for PROPKA functionality."""

from pathlib import Path

import common
import pytest


@pytest.mark.parametrize(
    "input_pdb", ["1K1I", "1AFS", "1FAS", "5DV8", "5D8V"], ids=str
)
def test_propka_apo(input_pdb, tmp_path):
    """PROPKA non-regression tests on biomolecules without ligands."""
    args = (
        "--log-level=INFO --ff=AMBER --drop-water "
        "--titration-state-method=propka"
    )
    output_pqr = Path(input_pdb).stem + ".pqr"
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize(
    "input_pdb", ["1K1I", "1AFS", "1FAS", "5DV8", "5D8V"], ids=str
)
def test_propka_pka(input_pdb):
    """PROPKA non-regression tests for pKa values on biomolecules without ligands."""
    output_csv = Path("tests/data") / f"{input_pdb}_pka.csv"
    common.run_propka_for_tests(
        input_pdb=input_pdb, compare_file=output_csv, pH=7.4
    )



"""Biomolecular residue class.

.. codeauthor:: Todd Dolinsky
.. codeauthor:: Nathan Baker
"""

import logging

from . import pdb, structures
from . import quatfit as quat
from . import utilities as util

_LOGGER = logging.getLogger(__name__)


class Residue:
    """Residue class

    .. todo::
       Should this class have a member variable for dihedrals? Pylint
       complains!

    The residue class contains a list of Atom objects associated with that
    residue and other helper functions.
    """

    def __init__(self, atoms: list[pdb.ATOM | pdb.HETATM]):
        """Initialize the class

        :param atoms:  list of atom-like (:class:`HETATM` or :class:`ATOM`)
            objects to be stored
        :type atoms:  list
        """
        sample_atom = atoms[-1]
        self.atoms: list[structures.Atom] = []
        self.name = sample_atom.res_name
        self.chain_id = sample_atom.chain_id
        self.res_seq = sample_atom.res_seq
        self.ins_code = sample_atom.ins_code
        self.map: dict[str, structures.Atom] = {}
        self.naname = None
        self.reference = None
        self.is_n_term = None
        self.is_c_term = None
        self.dihedrals = []
        atomclass = ""
        for atom in atoms:
            if isinstance(atom, pdb.ATOM):
                atomclass = "ATOM"
            elif isinstance(atom, pdb.HETATM):
                atomclass = "HETATM"
            atom = structures.Atom(atom, atomclass, self)
            atomname = atom.name
            if atomname not in self.map:
                self.add_atom(atom)
            else:  # Don't add duplicate atom
                oldatom = self.get_atom(atomname)
                oldatom.alt_loc = ""
        if self.name == "HOH":
            self.name = "WAT"
            for atom in self.atoms:
                atom.res_name = "WAT"

    def __str__(self):
        return f"{self.name} {self.chain_id} {self.res_seq}{self.ins_code}"

    def get_moveable_names(self, pivot):
        """Return all atom names that are further away than the pivot atom.

        :param residue:  the residue to use
        :type residue:  Residue
        :param pivot:  the pivot atomname
        :type pivot:  str
        :return:  names of atoms further away than pivot atom
        :rtype:  [str]
        """
        refdist = self.get_atom(pivot).refdistance
        return [atom.name for atom in self.atoms if atom.refdistance > refdist]

    def update_terminus_status(self):
        """Update the :makevar:`is_n_terms` and :makevar:`is_c_term` flags."""
        # If Nterm then update counter of hydrogens
        if self.is_n_term:
            count = 0
            atoms = ["H", "H2", "H3"]
            for atom in atoms:
                for atom2 in self.atoms:
                    atomname = atom2.name
                    if atom == atomname:
                        count += 1
            self.is_n_term = count
        # If Cterm then update counter
        if self.is_c_term:
            self.is_c_term = None
            for atom in self.atoms:
                atomname = atom.name
                if atomname == "HO":
                    self.is_c_term = 2
                    break
            if not self.is_c_term:
                self.is_c_term = 1

    def set_res_seq(self, value):
        """Change the residue sequence number.

        Set the atom field :makevar:`res_seq` and change the residue's
        information.
        The :makevar:`icode` field is no longer useful.

        :param value:  the new value of :makevar:`res_seq`
        :type value:  int
        """
        self.ins_code = ""
        self.res_seq = value
        for atom in self.atoms:
            atom.res_seq = value

    def set_chain_id(self, value):
        """Set the chain ID.

        :param value:  new :makevar:`chain_id` value
        :type value:  str
        """
        self.chain_id = value
        for atom in self.atoms:
            atom.chain_id = value

    def add_atom(self, atom):
        """Add the atom object to the residue.

        :param atom: atom-like object, e.g., :class:`HETATM` or :class:`ATOM`
        """
        self.atoms.append(atom)
        self.map[atom.name] = atom

    def remove_atom(self, atomname):
        """Remove an atom from the residue object.

        :param atomname:  the name of the atom to be removed
        :type atomname:  str
        """
        # Delete the atom from the map
        atom = self.map[atomname]
        bonds = atom.bonds
        del self.map[atomname]
        # Delete the atom from the list
        self.atoms.remove(atom)
        # Delete all instances of the atom as a bond
        for bondatom in bonds:
            if atom in bondatom.bonds:
                bondatom.bonds.remove(atom)
        del atom

    def rename_atom(self, oldname, newname):
        """Rename an atom to a new name.

        :param oldname:  old atom name
        :type oldname:  str
        :param newname:  new atom name
        :type newname:  str
        """
        atom = self.map[oldname]
        atom.name = newname
        self.map[newname] = atom
        del self.map[oldname]

    def get_atom(self, name: str) -> structures.Atom | None:
        """Retrieve a residue atom based on its name.

        :param resname:  name of the residue to retrieve
        :type resname:  str
        :return:  residue
        :rtype:  structures.Atom | None
        """
        return self.map.get(name)

    def has_atom(self, name):
        """Return True if atom in residue.

        :param name:  atom name in question
        :type name:  str
        :return:  True if atom in residue
        :rtype:  bool
        """
        return name in self.map

    @property
    def charge(self):
        """Get the total charge of the residue.

        In order to get rid of floating point rounding error, do a string
        transformation.

        Returns:
            charge: The charge of the residue (float)
        """
        charge = (atom.ffcharge for atom in self.atoms if atom.ffcharge)
        charge = sum(charge)
        charge = float(f"{charge:.4f}")
        return charge

    def rename_residue(self, name):
        """Rename the residue.

        :param name:  the new name of the residue
        :type name:  str
        """
        self.name = name
        for atom in self.atoms:
            atom.res_name = name

    @classmethod
    def rotate_tetrahedral(cls, atom1, atom2, angle):
        """Rotate about the atom1-atom2 bond by a given angle.

        All atoms connected to atom2 will rotate.

        :param atom1:  first atom of the bond to rotate about
        :type atom1:  structures.Atom
        :param atom2:  second atom of the bond to rotate about
        :type atom2:  structures.Atom
        :param angle:  degrees to rotate
        :type angle:  float
        """
        moveatoms = []
        movecoords = []
        initcoords = util.subtract(atom2.coords, atom1.coords)
        # Determine which atoms to rotate
        for atom in atom2.bonds:
            if atom == atom1:
                continue
            moveatoms.append(atom)
            movecoords.append(util.subtract(atom.coords, atom1.coords))
        newcoords = quat.qchichange(initcoords, movecoords, angle)
        for iatom, atom in enumerate(moveatoms):
            atom.x = newcoords[iatom][0] + atom1.x
            atom.y = newcoords[iatom][1] + atom1.y
            atom.z = newcoords[iatom][2] + atom1.z

    def pick_dihedral_angle(self, conflict_names, oldnum=None):
        """Choose an angle number to use in debumping.

        Instead of simply picking a random chiangle, this function
        uses a more intelligent method to improve efficiency.
        The algorithm uses the names of the conflicting atoms
        within the residue to determine which angle number
        has the best chance of fixing the problem(s). The method
        also insures that the same chiangle will not be run twice
        in a row.

        :param residue:  residue that is being debumped
        :type residue:  Residue
        :param conflict_names: list of atom names that are currently
            conflicts
        :type conflict_names: [str]
        :param oldnum:  old dihedral angle number
        :type oldnum:  int
        :return:  new dihedral angle number
        :rtype:  int
        """
        bestnum = -1
        best = 0
        ilist = list(range(len(self.dihedrals)))
        if oldnum is not None and oldnum >= 0 and ilist:
            del ilist[oldnum]
            test_dihedral_indices = ilist[oldnum:] + ilist[:oldnum]
        else:
            test_dihedral_indices = ilist
        for i in test_dihedral_indices:
            if i == oldnum:
                continue
            if self.dihedrals[i] is None:
                continue
            score = 0
            atomnames = self.reference.dihedrals[i].split()
            pivot = atomnames[2]
            moveablenames = self.get_moveable_names(pivot)
            if conflict_names == moveablenames:
                return i
            for name in conflict_names:
                if name in moveablenames:
                    score += 1
                    if score > best:
                        best = score
                        bestnum = i
        return bestnum

    def set_donors_acceptors(self):
        """Set the donors and acceptors within the residue."""
        if self.reference is None:
            return
        for atom in self.atoms:
            atomname = atom.name
            atom.hdonor = False
            atom.hacceptor = False
            if atomname.startswith("N"):
                bonded = 0
                for bondedatom in atom.bonds:
                    if bondedatom.is_hydrogen:
                        atom.hdonor = True
                        bonded = 1
                        break
                if not bonded and self.reference.name == "HIS":
                    atom.hacceptor = True

            elif atomname.startswith("O") or (
                atomname.startswith("S") and self.reference.name == "CYS"
            ):
                atom.hacceptor = True
                for bondedatom in atom.bonds:
                    if bondedatom.is_hydrogen:
                        atom.hdonor = True
                        break

    def reorder(self):
        """Reorder the atoms to start with N, CA, C, O if they exist."""
        templist = []
        if self.has_atom("N"):
            templist.append(self.get_atom("N"))
        if self.has_atom("CA"):
            templist.append(self.get_atom("CA"))
        if self.has_atom("C"):
            templist.append(self.get_atom("C"))
        if self.has_atom("O"):
            templist.append(self.get_atom("O"))
        # Add remaining atoms
        for atom in self.atoms:
            if atom.name not in ["N", "CA", "C", "O"]:
                templist.append(atom)
        # Change the list pointer
        self.atoms = templist[:]

    def letter_code(self) -> str:
        """Default letter code for residue.

        :return:  letter code for residue
        :rtype:  str
        """
        return "X"



"""Tests for ligand functionality."""

import logging
from math import isclose
from pathlib import Path

import common
import pandas as pd
import pytest
from ligand_results import (
    CHARGES_1HPX,
    FORMAL_CHARGE_RESULTS,
    PARAMETER_RESULTS,
    RING_RESULTS,
    TORSION_RESULTS,
)
from numpy.testing import assert_almost_equal

from pdb2pqr.ligand.mol2 import Mol2Molecule

_LOGGER = logging.getLogger(__name__)


ALL_LIGANDS = set(TORSION_RESULTS) | set(RING_RESULTS)
ALL_LIGANDS |= {
    "1HPX-ligand.mol2",
    "1QBS-ligand.mol2",
    "1US0-ligand.mol2",
    "adp.mol2",
    "acetate.mol2",
}
ALL_LIGANDS = sorted(ALL_LIGANDS)


def test_peoe_charges():
    """Specifically test PEOE charges."""
    ligand = Mol2Molecule()
    mol2_path = Path("tests/data/1HPX-ligand.mol2")
    with open(mol2_path) as mol2_file:
        ligand.read(mol2_file)
    # WARNING: Nathan? The old_total_charge is never used
    # old_total_charge = sum(
    #    atom.formal_charge for atom in ligand.atoms.values()
    # )
    ligand.assign_parameters()
    new_total_charge = 0
    test_charges = []
    for atom in ligand.atoms.values():
        new_total_charge += atom.charge
        test_charges.append(atom.charge)
    assert_almost_equal(test_charges, CHARGES_1HPX, verbose=True)


@pytest.mark.parametrize("input_mol2", ALL_LIGANDS)
def test_assign_parameters(input_mol2):
    """Tests to break basic ligand functionality."""
    ligand = Mol2Molecule()
    mol2_path = Path("tests/data") / input_mol2
    with open(mol2_path) as mol2_file:
        ligand.read(mol2_file)
    old_total_charge = sum(
        atom.formal_charge for atom in ligand.atoms.values()
    )
    ligand.assign_parameters()
    new_total_charge = 0
    test_results = []
    for atom in ligand.atoms.values():
        test_row = {
            "name": atom.name,
            "type": atom.type,
            "charge": atom.charge,
            "radius": atom.radius,
        }
        test_results.append(test_row)
        new_total_charge += atom.charge
    _LOGGER.debug(f"Test results: {test_results}")
    test_results = pd.DataFrame(test_results)
    test_results = test_results.set_index("name")
    # _LOGGER.debug(f"Test results:\n{test_results.to_string()}")
    _LOGGER.info(
        f"Total charge: {old_total_charge:5.2f} -> {new_total_charge:5.2f}"
    )
    expected_results = pd.DataFrame(PARAMETER_RESULTS[input_mol2])
    expected_results = expected_results.set_index("name")
    assert_almost_equal(
        test_results["charge"].to_numpy(),
        expected_results["charge"].to_numpy(),
    )
    assert_almost_equal(
        test_results["radius"].to_numpy(),
        expected_results["radius"].to_numpy(),
        verbose=True,
    )


@pytest.mark.parametrize("input_mol2", ALL_LIGANDS)
def test_formal_charge(input_mol2):
    """Testing formal charge calculation."""
    ligand = Mol2Molecule()
    mol2_path = Path("tests/data") / input_mol2
    with open(mol2_path) as mol2_file:
        ligand.read(mol2_file)
    expected_results = FORMAL_CHARGE_RESULTS[input_mol2]
    errors = []
    for iatom, atom in enumerate(ligand.atoms.values()):
        try:
            expected_charge = expected_results[iatom]
        except IndexError as e:
            err = f"Missing result for {atom.name}, {atom.type}, "
            err += f"{atom.formal_charge}"
            raise IndexError(err) from e
        if not isclose(atom.formal_charge, expected_charge):
            err = (
                f"Atom {atom.name} {atom.type} with bond order "
                f"{atom.bond_order}: expected {expected_charge}, "
                f"got {atom.formal_charge}"
            )
            errors.append(err)
    if errors:
        err = "Errors in test values:\n"
        err += "\n".join(errors)
        raise ValueError(err)


@pytest.mark.parametrize("input_mol2", TORSION_RESULTS)
def test_torsions(input_mol2):
    """Test assignment of torsion angles."""
    ligand = Mol2Molecule()
    mol2_path = Path("tests/data") / input_mol2
    with open(mol2_path) as mol2_file:
        ligand.read(mol2_file)
    test = set()
    # Only test heavy-atom torsions
    for torsion in ligand.torsions:
        has_hydrogen = any(atom.startswith("H") for atom in torsion)
        if not has_hydrogen:
            test.add(torsion)
    try:
        expected = TORSION_RESULTS[input_mol2]
        diff = test ^ expected
        if len(diff) > 0:
            err = (
                f"Torsion test failed for {input_mol2}:\n"
                f"Got: {sorted(test)}\n"
                f"Expected: {sorted(expected)}\n"
                f"Difference: {sorted(diff)}"
            )
            raise ValueError(err)
    except KeyError as e:
        err = f"No results for {input_mol2}: {sorted(ligand.torsions)}"
        raise KeyError(err) from e


@pytest.mark.parametrize("input_mol2", ALL_LIGANDS)
def test_rings(input_mol2):
    """Test assignment of torsion angles."""
    ligand = Mol2Molecule()
    mol2_path = Path("tests/data") / input_mol2
    with open(mol2_path) as mol2_file:
        ligand.read(mol2_file)
    test = ligand.rings
    try:
        benchmark = RING_RESULTS[input_mol2]
    except KeyError as e:
        err = f"Missing expected results for {input_mol2}: {test}"
        raise KeyError(err) from e
    diff = test ^ benchmark
    if len(diff) > 0:
        err = (
            f"Ring test failed for {input_mol2}:\n"
            f"Got: {sorted(test)}\n"
            f"Expected: {sorted(benchmark)}\n"
            f"Difference: {sorted(diff)}"
        )
        raise ValueError(err)
    for atom_name in ligand.atoms:
        atom = ligand.atoms[atom_name]
        if atom.num_rings > 0:
            str_ = f"{atom.num_rings} rings: {atom}"
            _LOGGER.debug(str_)


@pytest.mark.parametrize("input_pdb", ["1HPX"], ids=str)
def test_ligand_biomolecule(input_pdb, tmp_path):
    """PROPKA non-regression tests on biomolecules without ligands."""
    input_pdb = Path(input_pdb)
    ligand = Path("tests/data") / f"{input_pdb.stem}-ligand.mol2"
    args = f"--log-level=INFO --ff=AMBER --drop-water --ligand={ligand}"
    output_pqr = Path(input_pdb).stem + ".pqr"
    _LOGGER.debug(f"Running test in {tmp_path}")
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        tmp_path=tmp_path,
    )



"""Ligand topology classes."""

import logging

_LOGGER = logging.getLogger(__name__)


class Topology:
    """Ligand topology class."""

    def __init__(self, molecule):
        """Initialize with molecule.

        :param molecule:  Mol2Molecule object
        :type molecule:  Mol2Molecule
        """
        self.atom_dict = {}
        for atom in molecule.atoms:
            self.atom_dict[atom.name] = atom
            raise NotImplementedError



"""Perform functions related to _main_ execution of PDB2PQR.

This module is intended for functions that directly touch arguments provided at
the invocation of PDB2PQR.
It was created to avoid cluttering the __init__.py file.

.. codeauthor:: Nathan Baker (et al.)
"""

import argparse
import logging
import sys
import tempfile
from collections import OrderedDict
from collections.abc import Sequence
from io import StringIO
from os import PathLike
from pathlib import Path

import propka.input as pk_in
import propka.lib
import propka.output as pk_out
from propka.molecular_container import MolecularContainer
from propka.parameters import Parameters

from . import aa, debump, forcefield, hydrogens, io
from . import biomolecule as biomol
from .config import (
    CITATIONS,
    FORCE_FIELDS,
    IGNORED_PROPKA_OPTIONS,
    REPAIR_LIMIT,
    TITLE_STR,
    VERSION,
)
from .ligand.mol2 import Mol2Molecule
from .utilities import noninteger_charge

_LOGGER = logging.getLogger(f"PDB2PQR{VERSION}")


def build_main_parser():
    """Build an argument parser.

    .. todo::
        Need separate argparse groups for PDB2PKA and PROPKA.
        These exist but need real options.

    :returns:  argument parser
    :rtype:  argparse.ArgumentParser
    """
    desc = TITLE_STR
    pars = argparse.ArgumentParser(
        prog="pdb2pqr",
        description=desc,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        conflict_handler="resolve",
    )
    pars.add_argument(
        "input_path",
        help="Input PDB path or ID (to be retrieved from RCSB database",
    )
    pars.add_argument("output_pqr", help="Output PQR path")
    pars.add_argument(
        "--log-level",
        help="Logging level",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    grp1 = pars.add_argument_group(
        title="Mandatory options",
        description="One of the following options must be used",
    )
    grp1.add_argument(
        "--ff",
        choices=[ff.upper() for ff in FORCE_FIELDS],
        default="PARSE",
        help="The forcefield to use.",
    )
    grp1.add_argument(
        "--userff",
        help=(
            "The user-created forcefield file to use. Requires "
            "--usernames and overrides --ff"
        ),
    )
    grp1.add_argument(
        "--clean",
        action="store_true",
        default=False,
        help=(
            "Do no optimization, atom addition, or parameter assignment, "
            "just return the original PDB file in aligned format. Overrides "
            "--ff and --userff"
        ),
    )
    grp2 = pars.add_argument_group(title="General options")
    grp2.add_argument(
        "--nodebump",
        dest="debump",
        action="store_false",
        default=True,
        help="Do not perform the debumping operation",
    )
    grp2.add_argument(
        "--noopt",
        dest="opt",
        action="store_false",
        default=True,
        help="Do not perform hydrogen optimization",
    )
    grp2.add_argument(
        "--keep-chain",
        action="store_true",
        default=False,
        help="Keep the chain ID in the output PQR file",
    )
    grp2.add_argument(
        "--assign-only",
        action="store_true",
        default=False,
        help=(
            "Only assign charges and radii - do not add atoms, "
            "debump, or optimize."
        ),
    )
    grp2.add_argument(
        "--ffout",
        choices=[ff.upper() for ff in FORCE_FIELDS],
        help=(
            "Instead of using the standard canonical naming scheme for "
            "residue and atom names, use the names from the given forcefield"
        ),
    )
    grp2.add_argument(
        "--usernames",
        help=(
            "The user-created names file to use. Required if using --userff"
        ),
    )
    grp2.add_argument(
        "--apbs-input",
        help=(
            "Create a template APBS input file based on the generated PQR "
            "file at the specified location."
        ),
    )
    grp2.add_argument(
        "--pdb-output",
        default=None,
        help=(
            "Create a PDB file based on input. This will be missing charges "
            "and radii"
        ),
    )
    grp2.add_argument(
        "--ligand",
        help=(
            "Calculate the parameters for a single MOL2-format ligand at the "
            "path specified by this option.  The MOL2 ligand name should "
            "match only one ligand in the PDB file."
        ),
    )
    grp2.add_argument(
        "--whitespace",
        action="store_true",
        default=False,
        help=(
            "Insert whitespaces between atom name and residue name, between x "
            "and y, and between y and z."
        ),
    )
    grp2.add_argument(
        "--neutraln",
        action="store_true",
        default=False,
        help=(
            "Make the N-terminus of a protein neutral (default is "
            "charged). Requires PARSE force field."
        ),
    )
    grp2.add_argument(
        "--neutralc",
        action="store_true",
        default=False,
        help=(
            "Make the C-terminus of a protein neutral (default is "
            "charged). Requires PARSE force field."
        ),
    )
    grp2.add_argument(
        "--drop-water",
        action="store_true",
        default=False,
        help="Drop waters before processing biomolecule.",
    )
    grp2.add_argument(
        "--include-header",
        action="store_true",
        default=False,
        help=(
            "Include pdb header in pqr file. WARNING: The resulting PQR file "
            "will not work with APBS versions prior to 1.5"
        ),
    )
    grp3 = pars.add_argument_group(
        title="pKa options", description="Options for titration calculations"
    )
    grp3.add_argument(
        "--titration-state-method",
        dest="pka_method",
        choices=(["propka", "pkaani"]),
        help=(
            "Method used to calculate titration states. If a titration state "
            "method is selected, titratable residue charge states will be set "
            "by the pH value supplied by --with_ph. pKa-ANI uses a learned quantum mechanical"
            "representation of a residue's atomic environment from the ANI-2x neural network potential, and their pKa prediction model has been shown to surpass"
            "the accuracy of PROPKA (https://doi.org/10.1039/D1SC05610G). Incidentally, the paper shows that even a null model (simply guessing the mean of the test"
            "set repeatedly for each test case) also outperforms PROPKA (Figure 4)."
        ),
    )
    grp3.add_argument(
        "--with-ph",
        dest="ph",
        type=float,
        action="store",
        default=7.0,
        help=(
            "pH values to use when applying the results of the selected pH "
            "calculation method."
        ),
    )
    pars: argparse.ArgumentParser = propka.lib.build_parser(pars)

    # Override version flag set by PROPKA
    pars.add_argument(
        "--version", action="version", version=f"%(prog)s {VERSION}"
    )

    return pars


def print_splash_screen(args):
    """Print argument overview and citation information.

    :param args:  command-line arguments
    :type args:  argparse.Namespace
    """
    _LOGGER.debug(f"Args:  {args}")
    _LOGGER.info(TITLE_STR)
    for citation in CITATIONS:
        _LOGGER.info(citation)


def check_files(args):
    """Check for other necessary files.

    :param args:  command-line arguments
    :type args:  argparse.Namespace
    :raises FileNotFoundError:  necessary files not found
    :raises RuntimeError:  input argument or file parsing problems
    """
    if args.usernames is not None:
        usernames = Path(args.usernames)
        if not usernames.is_file():
            error = f"User-provided names file does not exist: {usernames}"
            raise FileNotFoundError(error)
    if args.userff is not None:
        userff = Path(args.userff)
        if not userff.is_file():
            error = f"User-provided forcefield file does not exist: {userff}"
            raise FileNotFoundError(error)
        if args.usernames is None:
            err = "--usernames must be specified if using --userff"
            raise RuntimeError(err)
    elif args.ff is not None:
        io.test_dat_file(args.ff)
    if args.ligand is not None:
        ligand = Path(args.ligand)
        if not ligand.is_file():
            error = f"Unable to find ligand file: {ligand}"
            raise FileNotFoundError(error)


def check_options(args):
    """Sanity check options.

    :param args:  command-line arguments
    :type args:  argparse.Namespace
    :raises RuntimeError:  silly option combinations were encountered.
    """
    for option, new_value in IGNORED_PROPKA_OPTIONS.items():
        if option in args:
            value = getattr(args, option)
            if value:
                _LOGGER.warning(
                    f"PROPKA option '{option}' {getattr(args, option)} is not "
                    f"processed correctly by PDB2PQR. Ignoring."
                )
                setattr(args, option, new_value)
    if (args.ph < 0) or (args.ph > 14):
        err = (
            f"Specified pH ({args.ph}) is outside the range "
            "[1, 14] of this program"
        )
        raise RuntimeError(err)
    if args.neutraln and (args.ff is None or args.ff.lower() != "parse"):
        err = "--neutraln option only works with PARSE forcefield!"
        raise RuntimeError(err)
    if args.neutralc and (args.ff is None or args.ff.lower() != "parse"):
        err = "--neutralc option only works with PARSE forcefield!"
        raise RuntimeError(err)


def print_pqr(args, pqr_lines, header_lines, missing_lines, is_cif):
    """Print PQR-format output to specified file

    .. todo::  Move this to another module (io)

    :param argparse.Namespace args:  command-line arguments
    :param [str] pqr_lines:  output lines (records)
    :param [str] header_lines:  header lines
    :param [str] missing_lines:  lines describing missing atoms (should go
        in header)
    :param bool is_cif:  flag indicating CIF format
    """
    with open(args.output_pqr, "w") as outfile:
        # Adding whitespaces if --whitespace is in the options
        if header_lines:
            _LOGGER.warning(
                f"Ignoring {len(header_lines)} header lines in output."
            )
        if missing_lines:
            _LOGGER.warning(
                f"Ignoring {len(missing_lines)} missing lines in output."
            )
        for line in pqr_lines:
            if args.whitespace:
                if line[0:4] == "ATOM" or line[0:6] == "HETATM":
                    newline = (
                        line[0:6]
                        + " "
                        + line[6:16]
                        + " "
                        + line[16:38]
                        + " "
                        + line[38:46]
                        + " "
                        + line[46:]
                    )
                    outfile.write(newline)
            elif line[0:3] != "TER" or not is_cif:
                outfile.write(line)
        if is_cif:
            outfile.write("#\n")


def print_pdb(args, pdb_lines, header_lines, missing_lines, is_cif):
    """Print PDB-format output to specified file

    .. todo::  Move this to another module (io)

    :param argparse.Namespace args:  command-line arguments
    :param [str]] pdb_lines:  output lines (records)
    :param [str] header_lines:  header lines
    :param [str] missing_lines:  lines describing missing atoms (should go in
        header)
    :param bool is_cif:  flag indicating CIF format
    """
    with open(args.pdb_output, "w") as outfile:
        # Adding whitespaces if --whitespace is in the options
        if header_lines:
            _LOGGER.warning(
                f"Ignoring {len(header_lines)} header lines in output."
            )
        if missing_lines:
            _LOGGER.warning(
                f"Ignoring {len(missing_lines)} missing lines in output."
            )
        for line in pdb_lines:
            if line[0:3] != "TER" or not is_cif:
                outfile.write(line)


def transform_arguments(args):
    """Transform arguments with logic not provided by argparse.

    .. todo::  I wish this could be done with argparse.

    :param args:  command-line arguments
    :type args:  argparse.Namespace
    :return:  modified arguments
    :rtype:  argparse.Namespace
    """
    if args.assign_only or args.clean:
        args.debump = False
        args.opt = False
    if args.userff is not None:
        args.userff = args.userff
    elif args.ff is not None:
        args.ff = args.ff.lower()
    if args.ffout is not None:
        args.ffout = args.ffout.lower()
    return args


def setup_molecule(pdblist, definition, ligand_path):
    """Set up the molecular system.

    :param pdblist:  list of PDB records
    :type pdblist:  list
    :param definition:  topology definition
    :type definition:  Definition
    :param ligand_path:  path to ligand (may be None)
    :type ligand_path:  str
    :return: (biomolecule object, definition object--revised if ligand was
        parsed, ligand object--may be None)
    :rtype: (Biomolecule, Definition, Ligand)
    """
    if ligand_path is not None:
        ligand = Mol2Molecule()
        with open(ligand_path, encoding="utf-8") as ligand_file:
            ligand.read(ligand_file)
    else:
        ligand = None
    biomolecule = biomol.Biomolecule(pdblist, definition)
    _LOGGER.info(
        f"Created biomolecule object with {len(biomolecule.residues)} "
        f"residues and {len(biomolecule.atoms)} atoms."
    )
    for residue in biomolecule.residues:
        multoccupancy = False
        for atom in residue.atoms:
            if atom.alt_loc != "":
                multoccupancy = True
                txt = f"Multiple occupancies found: {atom.name} in {residue}."
                _LOGGER.warning(txt)
        if multoccupancy:
            err = (
                f"Multiple occupancies found in {residue}. At least one of "
                "the instances is being ignored."
            )
            _LOGGER.warning(err)
    return biomolecule, definition, ligand


def is_repairable(biomolecule, has_ligand):
    """Determine if the biomolecule can be (or needs to be) repaired.

    :param biomolecule:  biomolecule object
    :type biomolecule:  biomol.Biomolecule
    :param has_ligand:  does the system contain a ligand?
    :type has_ligand:  bool
    :return:  indication of whether biomolecule can be repaired
    :rtype:  bool
    :raises ValueError: if there are insufficient heavy atoms or a significant
        part of the biomolecule is missing
    """
    num_heavy = biomolecule.num_heavy
    num_missing = biomolecule.num_missing_heavy
    if num_heavy == 0:
        if not has_ligand:
            err = (
                "No biomolecule heavy atoms found and no ligand present. "
                "Unable to proceed.  You may also see this message if "
                "PDB2PQR does not have parameters for any residue in your "
                "biomolecule."
            )
            raise ValueError(err)
        else:
            err = (
                "No heavy atoms found but a ligand is present. Proceeding "
                "with caution."
            )
            _LOGGER.warning(err)
            return False
    if num_missing == 0:
        _LOGGER.info("This biomolecule is clean.  No repair needed.")
        return False
    miss_frac = float(num_missing) / float(num_heavy)
    if miss_frac > REPAIR_LIMIT:
        error = f"This PDB file is missing too many ({num_missing} out of "
        error += f"{num_heavy:d}, {miss_frac:g}) "
        error += "heavy atoms to accurately repair the file."
        error += f"The current repair limit is set at {REPAIR_LIMIT:g}. "
        error += "You may also see this message if PDB2PQR does not have "
        error += "parameters for enough residues in your biomolecule."
        _LOGGER.error(error)
        return False
    return True


def drop_water(pdblist):
    """Drop waters from a list of PDB records.

    .. todo:: this module is already too long but this function fits better
        here. Other possible place would be utilities.

    :param pdb_list:  list of PDB records as returned by io.get_molecule
    :type pdb_list:  [str]
    :return:  new list of PDB records with waters removed.
    :rtype:  [str]
    """
    pdblist_new = []
    for record in pdblist:
        record_type = record.record_type()
        if (
            record_type in ["HETATM", "ATOM", "SIGATM", "SEQADV"]
            and record.res_name in aa.WAT.water_residue_names
        ):
            continue
        pdblist_new.append(record)
    return pdblist_new


def run_propka(args, biomolecule):
    """Run a PROPKA calculation.

    :param args:  command-line arguments
    :type args:  argparse.Namespace
    :param biomolecule:  biomolecule object
    :type biomolecule:  Biomolecule
    :return:  (DataFrame-convertible table of assigned pKa values,
               pKa information from PROPKA)
    :rtype:  (list, str)
    """
    lines = io.print_biomolecule_atoms(
        atomlist=biomolecule.atoms, chainflag=args.keep_chain, pdbfile=True
    )

    with StringIO() as fpdb:
        fpdb.writelines(lines)
        parameters = pk_in.read_parameter_file(args.parameters, Parameters())
        molecule = MolecularContainer(parameters, args)
        # needs a mock name with .pdb extension to work with stream data, hence the "input.pdb"
        molecule = pk_in.read_molecule_file("input.pdb", molecule, fpdb)

    molecule.calculate_pka()

    # Extract pKa information from PROPKA
    # write_pka(
    #    self,
    #    self.version.parameters,
    #    filename=filename,
    #    conformation="AVR",
    #    reference=reference,
    # )
    lines = []
    # lines.append(f"{pk_out.get_propka_header()}")
    # lines.append(f"{pk_out.get_references_header()}")
    # lines.append(
    #     pk_out.get_determinant_section(
    #         molecule, 'AVR', molecule.version.parameters
    #     )
    # )
    # lines.append(
    #     pk_out.get_summary_section(
    #         molecule, "AVR", molecule.version.parameters
    #     )
    # )
    # lines.append(pk_out.get_the_line())
    lines.append(
        pk_out.get_folding_profile_section(
            molecule,
            conformation="AVR",
            reference="neutral",
            window=[0.0, 14.0, 1.0],
        )
    )
    lines.append(
        pk_out.get_charge_profile_section(molecule, conformation="AVR")
    )
    lines.append(pk_out.get_the_line())
    pka_str = "\n".join(lines)

    # Summarize in pKas in DataFrame for later use
    conformation = molecule.conformations["AVR"]
    rows = []
    for group in conformation.groups:
        row_dict = OrderedDict()
        atom = group.atom
        row_dict["res_num"] = atom.res_num
        row_dict["ins_code"] = atom.icode
        row_dict["res_name"] = atom.res_name
        row_dict["chain_id"] = atom.chain_id
        row_dict["group_label"] = group.label
        row_dict["group_type"] = getattr(group, "type", None)
        row_dict["pKa"] = group.pka_value
        row_dict["model_pKa"] = group.model_pka
        row_dict["buried"] = group.buried
        if group.coupled_titrating_group:
            row_dict["coupled_group"] = group.coupled_titrating_group.label
        else:
            row_dict["coupled_group"] = None
        rows.append(row_dict)
    return rows, pka_str


def run_pkaani(args, biomolecule):
    """Run a pKa-ANI calculation.

    :param args:  command-line arguments
    :type args:  argparse.Namespace
    :param biomolecule:  biomolecule object
    :type biomolecule:  Biomolecule
    :return:  (DataFrame-convertible table of assigned pKa values,
               pKa information from pKa-ANI)
    :rtype:  list of OrderedDicts
    """
    pkaani_citation = """
    Gokcan, H. and Isayev, O. (2022) \'Prediction of protein pKa with representation learning\', 
    Chemical Science, 13(8), pp. 2462–2474. doi:10.1039/d1sc05610g. 
    """
    try:
        import warnings

        warnings.filterwarnings(
            "ignore", category=UserWarning, message=".*cuaev not installed.*"
        )
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
            message=".*pkg_resources is deprecated as an API.*",
        )
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
            message=".*torchani.data will not be available.*",
        )

        from pkaani.pkaani import calculate_pka as calculate_pkaani
        from tabulate import tabulate

    except ImportError as e:
        raise ImportError(
            "Optional module 'pkaani' is not installed. "
            "You can install it with: pip install pdb2pqr[pkaani]"
        ) from e

    _LOGGER.info(
        "If using pKa-ANI for titration state assignment, please cite: %s\n",
        pkaani_citation,
    )

    # pKa-ANI has its own way of parsing PDBs, will just pass the PDB
    # file to that and have it handle the parsing and PDB calculation
    lines = io.print_biomolecule_atoms(
        atomlist=biomolecule.atoms, chainflag=args.keep_chain, pdbfile=True
    )

    with tempfile.NamedTemporaryFile(
        mode="w+", suffix=".pdb", delete=True
    ) as tmp:
        for line in lines:
            tmp.write(f"{line}")
        tmp.flush()
        tmp.seek(0)

        pka = calculate_pkaani([tmp.name])[tmp.name]
        rows = []
        rows_2d_array = []

        for key in pka:
            row_dict = OrderedDict()
            row_dict["res_num"] = key[1]
            row_dict["res_name"] = pka[key][0]
            row_dict["chain_id"] = key[0]
            row_dict["pKa"] = pka[key][1]
            rows.append(row_dict)
            rows_2d_array.append([key[1], pka[key][0], key[0], pka[key][1]])

        pkaani_output = tabulate(
            rows_2d_array,
            headers=["Res. Number", "Res. Name", "Chain ID", "pKa"],
            tablefmt="grid",
        )
        _LOGGER.info("pKa results from pKa-ANI:\n%s", pkaani_output)

    return rows


def non_trivial(args, biomolecule, ligand, definition, is_cif):
    """Perform a non-trivial PDB2PQR run.

    .. todo::
       These routines should be generalized to biomolecules; none of them are
       specific to biomolecules.

    :param args:  command-line arguments
    :type args:  argparse.Namespace
    :param biomolecule:  biomolecule
    :type biomolecule:  Biomolecule
    :param ligand:  ligand object or None
    :type ligand:  Mol2Molecule
    :param definition:  topology definition
    :type definition:  Definition
    :param is_cif:  indicates whether file is CIF format
    :type is_cif:  bool
    :raises ValueError:  for missing atoms that prevent debumping
    :return:  dictionary with results
    :rtype:  dict
    """
    _LOGGER.info("Loading forcefield.")
    forcefield_ = forcefield.Forcefield(
        args.ff, definition, args.userff, args.usernames
    )
    _LOGGER.info("Loading hydrogen topology definitions.")
    hydrogen_handler = hydrogens.create_handler()
    debumper = debump.Debump(biomolecule)
    pka_df = None
    if args.assign_only:
        # TODO - I don't understand why HIS needs to be set to HIP for
        # assign-only
        biomolecule.set_hip()
    else:
        if is_repairable(biomolecule, args.ligand is not None):
            _LOGGER.info(
                f"Attempting to repair {biomolecule.num_missing_heavy:d} "
                "missing atoms in biomolecule."
            )
            biomolecule.repair_heavy()
        _LOGGER.info("Updating disulfide bridges.")
        biomolecule.update_ss_bridges()
        if args.debump:
            _LOGGER.info("Debumping biomolecule.")
            try:
                debumper.debump_biomolecule()
            except ValueError as err:
                err = f"Unable to debump biomolecule. {err}"
                raise ValueError(err)
        if args.pka_method == "propka":
            _LOGGER.info("Assigning titration states with PROPKA.")
            biomolecule.remove_hydrogens()
            pka_df, pka_str = run_propka(args, biomolecule)
            _LOGGER.info(f"PROPKA information:\n{pka_str}")
            biomolecule.apply_pka_values(
                forcefield_.name,
                args.ph,
                {
                    f"{row['res_name']} {row['res_num']} {row['chain_id']}": row[
                        "pKa"
                    ]
                    for row in pka_df
                    if row["group_label"].startswith(row["res_name"])
                },
            )
        elif args.pka_method == "pkaani":
            _LOGGER.info("Assigning titration states with pKa-ANI.")
            biomolecule.remove_hydrogens()
            pka_df = run_pkaani(args, biomolecule)
            biomolecule.apply_pka_values(
                forcefield_.name,
                args.ph,
                # no need for if clause, pka_df necessarily contains only the titratable residues
                # as defined by pKa-ANI (HIS/ASP/GLU/TYR/LYS)
                {
                    f"{row['res_name']} {row['res_num']} {row['chain_id']}": row[
                        "pKa"
                    ]
                    for row in pka_df
                },
            )

        _LOGGER.info("Adding hydrogens to biomolecule.")
        biomolecule.add_hydrogens()
        if args.debump:
            _LOGGER.info("Debumping biomolecule (again).")
            debumper.debump_biomolecule()
        _LOGGER.info("Optimizing hydrogen bonds")
        hydrogen_routines = hydrogens.HydrogenRoutines(
            debumper, hydrogen_handler
        )
        if args.opt:
            hydrogen_routines.set_optimizeable_hydrogens()
            biomolecule.hold_residues(None)
            hydrogen_routines.initialize_full_optimization()
        else:
            hydrogen_routines.initialize_wat_optimization()
        hydrogen_routines.optimize_hydrogens()
        hydrogen_routines.cleanup()
    _LOGGER.info("Applying force field to biomolecule states.")
    biomolecule.set_states()
    matched_atoms, missing_atoms = biomolecule.apply_force_field(forcefield_)
    if args.ligand is not None:
        _LOGGER.info("Processing ligand.")
        _LOGGER.warning("Using ZAP9 forcefield for ligand radii.")
        ligand.assign_parameters()
        lig_atoms = []
        for residue in biomolecule.residues:
            tot_charge = 0
            for pdb_atom in residue.atoms:
                # Only check residues with HETATM
                if pdb_atom.type == "ATOM":
                    break
                try:
                    mol2_atom = ligand.atoms[pdb_atom.name]
                    pdb_atom.radius = mol2_atom.radius
                    pdb_atom.ffcharge = mol2_atom.charge
                    tot_charge += mol2_atom.charge
                    lig_atoms.append(pdb_atom)
                except KeyError:
                    err = (
                        f"Can't find HETATM {residue.name} {residue.res_seq} "
                        f"{pdb_atom.name} in MOL2 file"
                    )
                    _LOGGER.warning(err)
                    missing_atoms.append(pdb_atom)
        matched_atoms += lig_atoms
    total_charge = 0
    for residue in biomolecule.residues:
        charge = residue.charge
        charge_err = noninteger_charge(charge)
        if charge_err:
            _LOGGER.warning(
                f"Residue {residue} has non-integer charge:  {charge_err}"
            )
        total_charge += charge
    charge_err = noninteger_charge(total_charge)
    if charge_err:
        raise ValueError(charge_err)
    if args.ffout is not None:
        _LOGGER.info(f"Applying custom naming scheme ({args.ffout}).")
        if args.ffout != args.ff:
            name_scheme = forcefield.Forcefield(args.ffout, definition, None)
        else:
            name_scheme = forcefield_
        biomolecule.apply_name_scheme(name_scheme)
    _LOGGER.info("Regenerating headers.")
    reslist, charge = biomolecule.charge
    if is_cif:
        header = io.print_pqr_header_cif(
            missing_atoms,
            reslist,
            charge,
            args.ff,
            args.pka_method,
            args.ph,
            args.ffout,
            include_old_header=args.include_header,
        )
    else:
        header = io.print_pqr_header(
            biomolecule.pdblist,
            missing_atoms,
            reslist,
            charge,
            args.ff,
            args.pka_method,
            args.ph,
            args.ffout,
            include_old_header=args.include_header,
        )
    _LOGGER.info("Regenerating PDB lines.")
    lines = io.print_biomolecule_atoms(matched_atoms, args.keep_chain)
    return {
        "lines": lines,
        "header": header,
        "missed_residues": missing_atoms,
        "pka_df": pka_df,
    }


def main_driver(args: argparse.Namespace):
    """Main driver for running program from the command line.

    Validate inputs, launch PDB2PQR, handle output.

    :param args:  command-line arguments
    :type args:  argparse.Namespace
    """
    _LOGGER.debug(f"Invoked with arguments: {args}")
    print_splash_screen(args)
    _LOGGER.info("Checking and transforming input arguments.")
    args = transform_arguments(args)
    check_files(args)
    check_options(args)
    _LOGGER.info("Loading topology files.")
    definition = io.get_definitions()
    _LOGGER.info(f"Loading molecule: {args.input_path}")
    pdblist, is_cif = io.get_molecule(args.input_path)
    if args.drop_water:
        _LOGGER.info("Dropping water from structure.")
        pdblist = drop_water(pdblist)
    _LOGGER.info("Setting up molecule.")
    biomolecule, definition, ligand = setup_molecule(
        pdblist, definition, args.ligand
    )
    _LOGGER.info("Setting termini states for biomolecule chains.")
    biomolecule.set_termini(neutraln=args.neutraln, neutralc=args.neutralc)
    biomolecule.update_bonds()
    if args.clean:
        _LOGGER.info(
            "Arguments specified cleaning only; skipping remaining steps."
        )
        results = {
            "header": "",
            "missed_residues": None,
            "biomolecule": biomolecule,
            "lines": io.print_biomolecule_atoms(
                biomolecule.atoms, args.keep_chain
            ),
            "pka_df": None,
        }
    else:
        try:
            results = non_trivial(
                args=args,
                biomolecule=biomolecule,
                ligand=ligand,
                definition=definition,
                is_cif=is_cif,
            )
        except ValueError as err:
            _LOGGER.critical(err)
            _LOGGER.critical("Giving up.")
            raise RuntimeError from err
    print_pqr(
        args=args,
        pqr_lines=results["lines"],
        header_lines=results["header"],
        missing_lines=results["missed_residues"],
        is_cif=is_cif,
    )
    if args.pdb_output:
        print_pdb(
            args=args,
            pdb_lines=io.print_biomolecule_atoms(
                biomolecule.atoms, chainflag=args.keep_chain, pdbfile=True
            ),
            header_lines=results["header"],
            missing_lines=results["missed_residues"],
            is_cif=is_cif,
        )
    if args.apbs_input:
        io.dump_apbs(args.output_pqr, args.apbs_input)
    return results["missed_residues"], results["pka_df"], biomolecule


def main():
    """Hook for command-line usage."""
    parser = build_main_parser()
    args = parser.parse_args()
    io.setup_logger(args.output_pqr, args.log_level)
    if main_driver(args) == 1:
        sys.exit(1)


def run_pdb2pqr(args: Sequence[str | PathLike]):
    """Run PDB2PQR with a list of arguments.

    Logger is not set up so that it can be called multiple times.

    :param args:  list of command-line arguments
    :type args:  list
    :return:  results of PDB2PQR run
    :rtype:  tuple
    """
    args_strlist = [str(arg) for arg in args]
    parser = build_main_parser()
    args_parsed = parser.parse_args(args_strlist)
    return main_driver(args_parsed)


def dx_to_cube():
    """Convert DX file format to Cube file format.

    The OpenDX file format is defined at
    <https://www.idvbook.com/wp-content/uploads/2010/12/opendx.pdf` and the
    Cube file format is defined at
    <https://docs.chemaxon.com/display/Gaussian_Cube_format.html>.

    .. todo:: This function should be moved into the APBS code base.
    """
    desc = f"{TITLE_STR}\ndx2cube: converting OpenDX-format files to "
    desc += "Gaussian Cube format since (at least) 2015"
    parser = argparse.ArgumentParser(
        description=desc,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("dx_input", help="name of the dx_input file")
    parser.add_argument("pqr_input", help="name of the pqr_input file")
    parser.add_argument("output", help="name of the output file")
    parser.add_argument(
        "--log-level",
        help="set logging level",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    args = parser.parse_args()
    log_level = getattr(logging, args.log_level)
    logging.basicConfig(level=log_level)
    _LOGGER.debug(f"Got arguments: {args}", args)
    _LOGGER.info(f"Reading PQR from {args.pqr_input}...")
    with open(args.pqr_input) as pqr_file:
        atom_list = io.read_pqr(pqr_file)
    _LOGGER.info(f"Reading DX from {args.dx_input}...")
    with open(args.dx_input) as dx_file:
        dx_dict = io.read_dx(dx_file)
    _LOGGER.info(f"Writing Cube to {args.output}...")
    with open(args.output, "w") as cube_file:
        io.write_cube(cube_file, dx_dict, atom_list)



"""Hydrogen optimization module for PDB2PQR.

This is an module for hydrogen optimization routines.

.. todo::
    This module has too many lines and should be simplified.

.. codeauthor:: Todd Dolinsky
.. codeauthor:: Jens Erik Nielsen
.. codeauthor:: Yong Huang
.. codeauthor:: Nathan Baker
"""

__author__ = "Todd Dolinsky, Jens Erik Nielsen, Yong Huang, Nathan Baker"
import logging
from xml import sax

from .. import aa, cells, io
from .. import definitions as defns
from .. import quatfit as quat
from .. import utilities as util
from ..config import HYD_DEF_PATH
from . import structures
from .structures import (
    HydrogenConformation,
    HydrogenDefinition,
    HydrogenHandler,
    PotentialBond,
)

_LOGGER = logging.getLogger(__name__)
_LOGGER.addFilter(io.DuplicateFilter())


#: A dictionary of the titration states for residues
TITRATION_DICT = {
    "ASH1c": "1",
    "ASH1t": "2",
    "ASH2c": "3",
    "ASH2t": "4",
    "ASP": "0",
    "GLH1c": "1",
    "GLH1t": "2",
    "GLH2c": "3",
    "GLH2t": "4",
    "GLU": "0",
    "ARG0": "1+2+3+4",
    "ARG": "1+2+3+4+5",
    "LYS": "1",
    "LYS0": "0",
    "TYR": "1",
    "TYR-": "0",
    "HSD": "1",
    "HSE": "2",
    "HSP": "1+2",
    "H3": "1",
    "H2": "2",
    "H3+H2": "1+2",
    "CTR01c": "1",
    "CTR01t": "2",
    "CTR02c": "3",
    "CTR02t": "4",
    "CTR-": "0",
}


def create_handler(hyd_path=HYD_DEF_PATH):
    """Create and populate a hydrogen handler.

    :param hyd_def_file:  path to hydrogen definition file
    :type hyd_def_file:  string or pathlib.Path object
    :return: HydrogenHandler object
    :rtype: HydrogenHandler
    """
    handler = HydrogenHandler()
    hyd_path = io.test_dat_file(hyd_path)
    with open(hyd_path) as hyd_file:
        sax.make_parser()
        sax.parseString(hyd_file.read(), handler)
    return handler


class HydrogenRoutines:
    """The main routines for hydrogen optimization.

    .. todo::
       This class really needs to be refactored.
    """

    def __init__(self, debumper, handler):
        """Initialize object.

        :param debumper:  Debump object
        :type debumper:  debump.Debump
        :param handler:  HydrogenHandler object
        :type handler:  HydrogenHandler
        """
        self.debumper = debumper
        self.biomolecule = debumper.biomolecule
        self.optlist = []
        self.atomlist = []
        self.resmap = {}
        self.hydrodefs = []
        self.map = handler.map

    def switchstate(self, states, amb, state_id):
        """Switch a residue to a new state by first removing all hydrogens.

        :param states:  the list of states
        :type states:  list
        :param amb:  the amibiguity to switch
        :type amb:  tup
        :param state_id:  the state id to switch to
        :type state_id:  int
        """
        if states == "pKa":
            return self.pka_switchstate(amb, state_id)

        if state_id > len(states):
            raise IndexError("Invalid State ID!")

        # First Remove all Hs
        residue = amb.residue
        hdef = amb.hdef
        for conf in hdef.conformations:
            hname = conf.hname
            boundname = conf.boundatom
            if residue.get_atom(hname) is not None:
                _LOGGER.debug(
                    f"Removing {residue.name} {residue.res_seq} {hname}"
                )
                residue.remove_atom(hname)
            residue.get_atom(boundname).hacceptor = 1
            residue.get_atom(boundname).hdonor = 0

        # Update the IntraBonds
        name = residue.name
        defresidue = self.debumper.aadef.get_residue(name)
        residue.updateIntraBonds(defresidue)

        # Now build appropriate atoms
        state = states[state_id]
        for conf in state:
            _LOGGER.debug(conf)
            refcoords = []
            defcoords = []
            defatomcoords = []
            if conf == ():
                continue  # Nothing to add
            hname = conf.hname
            for atom in conf.atoms:
                # print confatoms
                atomname = atom.name
                resatom = residue.get_atom(atomname)
                if atomname == hname:
                    defatomcoords = atom.coords
                elif resatom is not None:
                    refcoords.append(resatom.coords)
                    defcoords.append(atom.coords)
                else:
                    raise KeyError("Could not find necessary atom!")

            newcoords = quat.find_coordinates(
                3, refcoords, defcoords, defatomcoords
            )
            boundname = conf.boundatom
            residue.create_atom(hname, newcoords, "ATOM")
            residue.addDebumpAtom(residue.get_atom(hname))
            residue.get_atom(boundname).addIntraBond(hname)
            residue.get_atom(boundname).hacceptor = 0
            residue.get_atom(boundname).hdonor = 1
            # Setting the SybylType for the newly built H
            residue.get_atom(hname).sybyl_type = "H"
            # formal charge for PEOE_PB
            residue.get_atom(hname).formalcharge = 0.0
            # flag the added hydrogen
            residue.get_atom(hname).titratableH = True
            residue.get_atom(hname).addIntraBond(boundname)

    @classmethod
    def pka_switchstate(cls, amb, state_id_):
        """Switch a residue to a new state by first removing all hydrogens.

        This routine is used in pKa calculations only!

        :param amb:  the amibiguity to switch
        :type amb:  tup
        :param state_id_:  the state id to switch to
        :type state_id_:  int
        """
        titrationdict = TITRATION_DICT
        state_id = titrationdict[state_id_]
        state_id = state_id.split("+")
        new_state_id = [int(i) for i in state_id]
        residue = amb.residue
        hdef = amb.hdef
        for conf in hdef.conformations:
            hname = conf.hname
            boundname = conf.boundatom
            if residue.get_atom(hname) is not None:
                residue.remove_atom(hname)
            residue.get_atom(boundname).hacceptor = 1
            residue.get_atom(boundname).hdonor = 0

        # Update the IntraBonds
        for state_id in new_state_id:
            if state_id == 0:
                continue
            conf = hdef.conformations[state_id - 1]
            refcoords = []
            defcoords = []
            defatomcoords = []
            if conf == ():
                continue
            hname = conf.hname
            for atom in conf.atoms:
                if (
                    residue.is_n_term
                    and residue.name == "PRO"
                    and atom.name == "H"
                ):
                    atom.name = "CD"
                    atom.x = 1.874
                    atom.y = 0.862
                    atom.z = 1.306

            if not residue.rebuild_tetrahedral(hname):
                for atom in conf.atoms:
                    atomname = atom.name
                    resatom = residue.get_atom(atomname)
                    if atomname == hname:
                        defatomcoords = atom.coords
                    elif resatom is not None:
                        refcoords.append(resatom.coords)
                        defcoords.append(atom.coords)
                    else:
                        raise KeyError("Could not find necessary atom!")

                newcoords = quat.find_coordinates(
                    3, refcoords, defcoords, defatomcoords
                )
                residue.create_atom(hname, newcoords)

            boundname = conf.boundatom
            residue.get_atom(boundname).hacceptor = 0
            residue.get_atom(boundname).hdonor = 1

            # Setting the SybylType for the newly built H
            residue.get_atom(hname).sybyl_type = "H"

            # formal charge for PEOE_PB
            residue.get_atom(hname).formalcharge = 0.0

            # flag the added hydrogen
            residue.get_atom(hname).titratableH = True

        # Update intrabonds again
        if residue.is_n_term and residue.name == "PRO":
            for atom in residue.atoms:
                if atom.name == "H":
                    residue.remove_atom("H")
        residue.update_terminus_status()

    def cleanup(self):
        """Delete extra carboxylic atoms.

        If there are any extra carboxlyic ``*1`` atoms, delete them.
        This may occur when no optimization is chosen.
        """
        for residue in self.debumper.biomolecule.residues:
            if not isinstance(residue, aa.Amino):
                continue
            if residue.name == "GLH" or "GLH" in residue.patches:
                if residue.has_atom("HE1") and residue.has_atom("HE2"):
                    residue.remove_atom("HE1")
            elif residue.name == "ASH" or "ASH" in residue.patches:
                if residue.has_atom("HD1") and residue.has_atom("HD2"):
                    residue.remove_atom("HD1")

    def is_optimizeable(self, residue):
        """Check to see if the given residue is optimizeable.

        There are three ways to identify a residue:

        1. By name (i.e., HIS)
        2. By reference name - a PDB file HSP has a HIS reference name
        3. By patch - applied by :program:`propka`, terminal selection

        :param residue:  the residue in question
        :type residue:  Residue
        :return: None if not optimizeable, otherwise the OptimizationHolder
            instance that corresponds to the residue.
        :rtype: None or OptimizationHolder
        """
        optinstance = None
        if not isinstance(residue, (aa.Amino, aa.WAT)):
            return optinstance

        if residue.name in self.map:
            optinstance = self.map[residue.name]
        elif residue.reference.name in self.map:
            optinstance = self.map[residue.reference.name]
        else:
            for patch in residue.patches:
                if patch in self.map:
                    optinstance = self.map[patch]
                    break

        # If alcoholic, make sure the hydrogen is present
        if optinstance is not None and optinstance.opttype == "Alcoholic":
            atomname = next(iter(optinstance.map.keys()))
            if not residue.reference.has_atom(atomname):
                optinstance = None
        return optinstance

    def set_optimizeable_hydrogens(self):
        """Set any hydrogen listed in HYDROGENS.xml that is optimizeable.

        Used BEFORE hydrogen optimization to label atoms so that they won't be
        debumped - i.e. if SER HG is too close to another atom, don't debump
        but wait for optimization.

        .. note::
           This function should not be used if full optimization is not taking
           place.
        """
        for residue in self.biomolecule.residues:
            optinstance = self.is_optimizeable(residue)
            if optinstance is None:
                continue
            for atom in residue.atoms:
                if atom.name in optinstance.map:
                    atom.optimizeable = 1

    def initialize_full_optimization(self):
        """Initialize the full optimization.

        Detects all optimizeable donors and acceptors and sets the internal
        optlist.
        """
        # Do some setup
        self.debumper.cells = cells.Cells(5)
        self.debumper.cells.assign_cells(self.biomolecule)
        self.biomolecule.calculate_dihedral_angles()
        self.biomolecule.set_donors_acceptors()
        self.biomolecule.update_internal_bonds()
        self.biomolecule.set_reference_distance()
        self.optlist = []
        self.atomlist = []
        # First initialize the various types
        for residue in self.biomolecule.residues:
            optinstance = self.is_optimizeable(residue)
            if isinstance(residue, aa.Amino):
                residue.fixed = (
                    1 if False in residue.stateboolean.values() else 0
                )
            if optinstance is None:
                continue

            type_ = optinstance.opttype
            if residue.fixed != 1:
                klass = getattr(structures, type_)
                myobj = klass(residue, optinstance, self.debumper)
                self.atomlist += myobj.atomlist
                self.optlist.append(myobj)
                self.resmap[residue] = myobj
        _LOGGER.debug("Done.")

    def initialize_wat_optimization(self):
        """Initialize optimization for waters only.

        Detects all optimizeable donors and acceptors and sets the internal
        optlist.
        """
        _LOGGER.info("Initializing water bonding optimization...")
        # Do some setup
        self.debumper.cells = cells.Cells(5)
        self.debumper.cells.assign_cells(self.biomolecule)
        self.biomolecule.calculate_dihedral_angles()
        self.biomolecule.set_donors_acceptors()
        self.biomolecule.update_internal_bonds()
        self.biomolecule.set_reference_distance()
        self.optlist = []
        # First initialize the various types
        for residue in self.biomolecule.residues:
            optinstance = self.is_optimizeable(residue)
            if optinstance is None:
                continue
            type_ = optinstance.opttype
            if type_ == "Water":
                klass = getattr(structures, type_)
                myobj = klass(residue, optinstance, self.debumper)
                self.atomlist += myobj.atomlist
                self.optlist.append(myobj)
                self.resmap[residue] = myobj
        _LOGGER.debug("Done.")

    def optimize_hydrogens(self):
        """The main driver for the optimization.

        .. note::
           Should be called only after the optlist has been initialized.

        .. todo::
           Remove hard-coded :makevar:`progress` threshold and increment
           values.

        .. todo::
           This function needs to be simplified.
        """
        _LOGGER.debug("Optimization progress:")
        optlist = self.optlist
        connectivity = {}
        # Initialize the detection progress
        if len(optlist) == 0:
            return
        _LOGGER.debug("  Detecting potential hydrogen bonds")
        progress = 0.0
        increment = 1.0 / len(optlist)
        for obj in optlist:
            connectivity[obj] = []
            for atom in obj.atomlist:
                closeatoms = self.debumper.cells.get_near_cells(atom)
                for closeatom in closeatoms:
                    # Conditions for continuing
                    if atom.residue == closeatom.residue:
                        continue
                    if not (closeatom.hacceptor or closeatom.hdonor):
                        continue
                    if (
                        atom.hdonor
                        and not atom.hacceptor
                        and not closeatom.hacceptor
                    ):
                        continue
                    if (
                        atom.hacceptor
                        and not atom.hdonor
                        and not closeatom.hdonor
                    ):
                        continue
                    dist = util.distance(atom.coords, closeatom.coords)
                    if dist < 4.3:
                        residue = atom.residue
                        hbond = PotentialBond(atom, closeatom, dist)
                        # Store the potential bond
                        obj.hbonds.append(hbond)
                        # Keep track of connectivity
                        if closeatom in self.atomlist:
                            closeobj = self.resmap[closeatom.residue]
                            if closeobj not in connectivity[obj]:
                                connectivity[obj].append(closeobj)
            progress += increment
            while progress >= 0.0499:
                progress -= 0.05
        # Some residues might have no nearby hbonds - if so, place at
        # default state
        for obj in optlist:
            if len(obj.hbonds) == 0:
                if obj.residue.fixed:
                    continue
                _LOGGER.debug(
                    f"{obj.residue} has no nearby partners - fixing."
                )
                obj.finalize()
        # Determine the distinct networks
        networks = []
        seen = []
        for obj1 in optlist:
            if obj1.residue.fixed:
                continue
            if obj1 in seen:
                continue
            network = util.analyze_connectivity(connectivity, obj1)
            for obj2 in network:
                if obj2 not in seen:
                    seen.append(obj2)
            networks.append(network)
        # Initialize the output progress
        if len(networks) > 0:
            _LOGGER.debug("Optimizing hydrogen bonds")
            progress = 0.0
            increment = 1.0 / len(networks)
        # Work on the networks
        for network in networks:
            txt = ""
            for obj in network:
                txt += f"{obj}, "
            _LOGGER.debug(f"Starting network {txt[:-2]}")
            #  FIRST:  Only optimizeable to backbone atoms
            _LOGGER.debug("* Optimizeable to backbone *")
            hbondmap = {}
            for obj in network:
                for hbond in obj.hbonds:
                    if hbond.atom2 not in self.atomlist:
                        hbondmap[hbond] = hbond.dist
            hbondlist = util.sort_dict_by_value(hbondmap)
            hbondlist.reverse()
            for hbond in hbondlist:
                atom = hbond.atom1
                atom2 = hbond.atom2
                obj = self.resmap[atom.residue]

                if atom.residue.fixed:
                    continue
                if atom.hdonor:
                    obj.try_donor(atom, atom2)
                if atom.hacceptor:
                    obj.try_acceptor(atom, atom2)
            # SECOND:  Non-dual water Optimizeable to Optimizeable
            _LOGGER.debug("* Optimizeable to optimizeable *")
            hbondmap = {}
            seenlist = []
            for obj in network:
                for hbond in obj.hbonds:
                    if hbond.atom2 in self.atomlist and not isinstance(
                        hbond.atom1.residue, aa.WAT
                    ):
                        if not isinstance(hbond.atom2.residue, aa.WAT):
                            # Only get one hbond pair
                            if (hbond.atom2, hbond.atom1) not in seenlist:
                                hbondmap[hbond] = hbond.dist
                                seenlist.append((hbond.atom1, hbond.atom2))
            hbondlist = util.sort_dict_by_value(hbondmap)
            hbondlist.reverse()
            for hbond in hbondlist:
                atom = hbond.atom1
                atom2 = hbond.atom2
                obj1 = self.resmap[atom.residue]
                obj2 = self.resmap[atom2.residue]
                # Atoms may no longer exist if already optimized
                if not atom.residue.has_atom(atom.name):
                    continue
                if not atom2.residue.has_atom(atom2.name):
                    continue
                res = 0
                if atom.hdonor and atom2.hacceptor:
                    res = obj1.try_both(atom, atom2, obj2)
                if atom.hacceptor and atom2.hdonor and res == 0:
                    obj2.try_both(atom2, atom, obj1)
            # THIRD:  All water-water residues
            _LOGGER.debug("* Water to Water *")
            hbondmap = {}
            seenlist = []
            for obj in network:
                for hbond in obj.hbonds:
                    residue = hbond.atom1.residue
                    if isinstance(residue, aa.WAT) and isinstance(
                        hbond.atom2.residue, aa.WAT
                    ):
                        if (hbond.atom2, hbond.atom1) not in seenlist:
                            hbondmap[hbond] = hbond.dist
                            seenlist.append((hbond.atom1, hbond.atom2))
            hbondlist = util.sort_dict_by_value(hbondmap)
            hbondlist.reverse()
            for hbond in hbondlist:
                atom = hbond.atom1
                atom2 = hbond.atom2
                obj1 = self.resmap[atom.residue]
                obj2 = self.resmap[atom2.residue]
                res = 0
                if atom.hdonor and atom2.hacceptor:
                    res = obj1.try_both(atom, atom2, obj2)
                if atom.hacceptor and atom2.hdonor and res == 0:
                    obj2.try_both(atom2, atom, obj1)
            # FOURTH: Complete all residues
            for obj in network:
                obj.complete()
            # STEP 5:  Update progress meter
            progress += 100.0 * increment
            while progress >= 5.0:
                progress -= 5.0

    def parse_hydrogen(self, res, topo):
        """Parse a list of lines in order to make a hydrogen definition.

        This is the current definition:
        ``Name Ttyp  A R # Stdconf HT Chi OPTm``

        .. todo::
           The type of the :makevar:`res` appears to be incorrect.

        .. todo::
           This function is too long and needs to be simplified.

        :param res:  the lines to parse (list)
        :type res:  unknown
        :param topo:  Topology object
        :type topo:  pdb2pqr.topology.Topology
        :return:  the hydrogen definition object
        :rtype: HydrogenDefinition

        """
        name = self.map[res].name
        opttype = self.map[res].opttype
        optangle = self.map[res].optangle
        map_ = self.map[res].map
        mydef = HydrogenDefinition(name, opttype, optangle, map_)
        patch_map = []
        atoms = []
        refatoms = []
        conformernames = []
        refmap = {}
        titrationstatemap = {}
        tautomermap = {}
        conformermap = {}
        atommap = {}
        ntrmap = {}
        hmap = {}
        nonhmap = {}
        # reference map from TOPOLOGY.xml
        for res_ in topo.residues:
            refmap[res_.name] = res_.reference
            for atom in refmap[res_.name].atoms:
                atommap[res_.name, atom.name] = atom
            for titrationstate in res_.titration_states:
                titrationstatemap[titrationstate.name] = titrationstate
                for tautomer in titrationstate.tautomers:
                    tautomermap[tautomer.name] = tautomer
                    for conformer in tautomer.conformers:
                        conformermap[conformer.name] = conformer
        if name == "CYS":
            _ = refmap["CYS"]
            atoms = ["HG"]
            refatoms = ["SG", "CB"]
        elif name == "HIS":
            _ = refmap["HIS"]
            atoms = ["HD1", "HE2"]
            for atom in atoms:
                refatoms = ["ND1", "CG", "CE1"]
        elif name == "LYS":
            _ = self.debumper.biomolecule.reference_map[name]
            patch_map = self.debumper.biomolecule.patch_map["LYN"]
            atoms = patch_map.remove
            refatoms = ["HZ1", "HZ2", "NZ"]
        elif name == "TYR":
            _ = self.debumper.biomolecule.reference_map[name]
            patch_map = self.debumper.biomolecule.patch_map["TYM"]
            atoms = patch_map.remove
            refatoms = ["OH", "CZ", "CE2"]
        elif name == "WAT":
            _ = self.debumper.biomolecule.reference_map[name]
            patch_map = self.debumper.biomolecule.patch_map["HOH"]
            atoms = ["H1", "H2"]
            refatoms = None
        elif name == "NTR":
            ntrmap = {}  # map for N-TERM
            for tautomer in titrationstatemap["NTER"].tautomers:
                for conformer in tautomermap[tautomer.name].conformers:
                    for conformeradds in conformermap[
                        conformer.name
                    ].conformer_adds:
                        for atom in conformeradds.atoms:
                            ntrmap[atom.name] = atom
            atoms = ["H3", "H2"]
            refatoms = ["CA", "H", "N"]
        elif name == "CTR":
            hmap = {}  # map for h atoms
            nonhmap = {}  # map for refatoms
            conformernames = []
            for tautomer in titrationstatemap["CTER"].tautomers:
                for conformer in tautomermap[tautomer.name].conformers:
                    for conformeradds in conformermap[
                        conformer.name
                    ].conformer_adds:
                        for atom in conformeradds.atoms:
                            nonhmap[atom.name] = atom
            for tautomer in titrationstatemap["CTER0"].tautomers:
                for conformer in tautomermap[tautomer.name].conformers:
                    conformernames.append(conformer.name)
                    for conformeradds in conformermap[
                        conformer.name
                    ].conformer_adds:
                        for atom in conformeradds.atoms:
                            hmap[conformer.name, atom.name] = atom
            atoms = ["HO"]
            refatoms = ["O", "C", "OXT"]
        elif name in ["SER", "GLN", "THR", "ARG", "ASN"]:
            _ = refmap[name]
            if name == "SER":
                atoms = ["HG"]
                refatoms = ["OG", "CB"]
            elif name == "GLN":
                atoms = ["HE21"]
                refatoms = ["NE2"]
            elif name == "THR":
                atoms = ["HG1"]
                refatoms = ["OG1", "CB"]
            elif name == "ARG":
                atoms = ["HH11", "HH12", "HH21", "HH22", "HE"]
                for atom in atoms:
                    refatoms = ["NH1", "NH2", "CZ"]
            elif name == "ASN":
                atoms = ["HD21"]
                refatoms = ["ND2"]
        elif name == "ASH":
            hmap = {}  # map for h atoms
            nonhmap = {}  # map for refatoms
            conformernames = []
            _ = refmap["ASP"]
            for tautomer in titrationstatemap["ASH"].tautomers:
                for conformer in tautomermap[tautomer.name].conformers:
                    for conformeradds in conformermap[
                        conformer.name
                    ].conformer_adds:
                        for atom in conformeradds.atoms:
                            hmap[conformer.name, atom.name] = atom
                            conformernames.append(conformer.name)
            atoms = ["HD1", "HD2"]
            refatoms = ["OD1", "CG", "OD2"]
        elif name == "GLH":
            hmap = {}  # map for h atoms
            nonhmap = {}  # map for refatoms
            conformernames = []
            _ = refmap["GLU"]
            for tautomer in titrationstatemap["GLH"].tautomers:
                for conformer in tautomermap[tautomer.name].conformers:
                    for conformeradds in conformermap[
                        conformer.name
                    ].conformer_adds:
                        for atom in conformeradds.atoms:
                            hmap[conformer.name, atom.name] = atom
                            conformernames.append(conformer.name)
            atoms = ["HE1", "HE2"]
            refatoms = ["OE1", "CD", "OE2"]
        else:
            patch_map = self.debumper.biomolecule.patch_map[name]
            atoms = list(patch_map.map.keys())
            atoms.sort()
        if name in ["NTR"]:
            bondlength = 1.0
            for atom in atoms:
                hname = atom
                x = ntrmap[hname].x
                y = ntrmap[hname].y
                z = ntrmap[hname].z
                bondatom = ntrmap[hname].bonds[0]
                myconf = HydrogenConformation(hname, bondatom, bondlength)
                atom = defns.DefinitionAtom(hname, x, y, z)
                myconf.add_atom(atom)

                # TODO - lots of arbitrary undefined numbers in this section
                for atom_ in refatoms:
                    if atom_ == "N":
                        natom = defns.DefinitionAtom(atom_, 1.201, 0.847, 0.0)
                        myconf.add_atom(natom)
                    elif atom_ == "CA":
                        caatom = defns.DefinitionAtom(atom_, 0.0, 0.0, 0.0)
                        myconf.add_atom(caatom)
                    elif atom_ == "H":
                        caatom = defns.DefinitionAtom(
                            atom_, 1.201, 1.847, 0.000
                        )
                        myconf.add_atom(caatom)
                mydef.add_conf(myconf)
        elif name in ["CTR"]:
            bondlength = 1.0
            for conformer in conformernames:
                for atom in atoms:
                    hname = atom
                    x = hmap[conformer, hname].x
                    y = hmap[conformer, hname].y
                    z = hmap[conformer, hname].z
                    bondatom = hmap[conformer, hname].bonds[0]
                    myconf = HydrogenConformation(hname, bondatom, bondlength)
                    atom = defns.DefinitionAtom(hname, x, y, z)
                    myconf.add_atom(atom)

                    # TODO - the following code is almost nonsensical
                    for atom_ in refatoms:
                        if atom_ == "C":
                            catom = defns.DefinitionAtom(
                                atom_, -1.250, 0.881, 0.000
                            )
                            myconf.add_atom(catom)
                        else:
                            atomname = atom_
                            x = nonhmap[atom_].x
                            y = nonhmap[atom_].y
                            z = nonhmap[atom_].z
                            atom2 = defns.DefinitionAtom(atomname, x, y, z)
                            myconf.add_atom(atom2)
                    mydef.add_conf(myconf)

        elif name in ["ASH", "GLH"]:
            for conformer in conformernames:
                for atom in atoms:
                    hname = atom
                    if ("1" in conformer and "1" in atom) or (
                        "2" in conformer and "2" in atom
                    ):
                        x = hmap[conformer, hname].x
                        y = hmap[conformer, hname].y
                        z = hmap[conformer, hname].z
                        bondatom = hmap[conformer, hname].bonds[0]
                        bondlength = 1.0
                        myconf = HydrogenConformation(
                            hname, bondatom, bondlength
                        )
                        atom = defns.DefinitionAtom(hname, x, y, z)
                        myconf.add_atom(atom)

                        for atom_ in refatoms:
                            atomname = atom_
                            refresname = ""
                            if name == "ASH":
                                refresname = "ASP"
                            elif name == "GLH":
                                refresname = "GLU"
                            x = atommap[refresname, atom_].x
                            y = atommap[refresname, atom_].y
                            z = atommap[refresname, atom_].z
                            atom2 = defns.DefinitionAtom(atomname, x, y, z)
                            myconf.add_atom(atom2)
                        mydef.add_conf(myconf)
        elif name not in ["WAT"]:
            bondlength = 1.0
            for atom in atoms:
                hname = atom
                x = atommap[name, hname].x
                y = atommap[name, hname].y
                z = atommap[name, hname].z
                bondatom = atommap[name, hname].bonds[0]
                myconf = HydrogenConformation(hname, bondatom, bondlength)
                atom = defns.DefinitionAtom(hname, x, y, z)
                myconf.add_atom(atom)
                if refatoms is not None:
                    if name == "HIS" and atom.name == "HE2":
                        refatoms = ["NE2", "CE1", "CD2"]
                    if name == "ARG" and atom.name == "HE":
                        refatoms = ["NE", "CZ", "NH1"]
                    # FIXME: 2020/07/06 intendo -  the "atom" is reused in
                    # the outer for loop and ambiguous
                    for atom in refatoms:
                        atomname = atom
                        x = atommap[name, atomname].x
                        y = atommap[name, atomname].y
                        z = atommap[name, atomname].z
                        atom = defns.DefinitionAtom(atomname, x, y, z)
                        myconf.add_atom(atom)
                    mydef.add_conf(myconf)
        return mydef

    def read_hydrogen_def(self, topo):
        """Read the hydrogen definition file

        :param topo:  Topology object
        :type topo:  Topology object
        """
        self.hydrodefs = []
        for mapping in self.map:
            res = mapping
            mydef = self.parse_hydrogen(res, topo)
            self.hydrodefs.append(mydef)
            res = ""



"""Hydrogen optimization routines.

.. codeauthor:: Todd Dolinsky
.. codeauthor:: Jens Erik Nielsen
.. codeauthor:: Yong Huang
.. codeauthor:: Nathan Baker
"""

import logging
import math

from .. import quatfit as quat
from .. import utilities as util
from ..config import ANGLE_CUTOFF, DIST_CUTOFF

_LOGGER = logging.getLogger(__name__)


class Optimize:
    """The holder class for the hydrogen optimization routines.

    Individual optimization types inherit off of this class.
    Any functions used by multiple types appear here.
    """

    def __init__(self):
        self.residue = None
        self.optinstance = None
        self.routines = None

    def __str__(self):
        return f"{self.residue} ({self.optinstance.opttype})"

    @staticmethod
    def get_hbond_angle(atom1, atom2, atom3):
        """Get the angle between three atoms

        :param atom1:  the first atom
        :type atom1:  Atom
        :param atom2:  the second (vertex) atom
        :type atom2:  Atom
        :param atom3:  the third atom
        :type atom3:  Atom
        :return:  the angle between the atoms in degrees
        :rtype:  float
        """
        atom_coords = atom2.coords
        coords1 = util.subtract(atom3.coords, atom_coords)
        coords2 = util.subtract(atom1.coords, atom_coords)
        norm1 = util.normalize(coords1)
        norm2 = util.normalize(coords2)
        dotted = util.dot(norm1, norm2)
        if dotted > 1.0:  # If normalized, this is due to rounding error
            dotted = 1.0
        elif dotted < -1.0:  # If normalized, this is due to rounding error
            dotted = -1.0
        rad = abs(math.acos(dotted))
        angle = rad * 180.0 / math.pi
        if angle > 180.0:
            angle = 360.0 - angle
        return angle

    def is_hbond(self, donor, acc):
        """Determine whether this donor acceptor pair is a hydrogen bond.

        .. todo::
           Remove hard-coded hydrogen bond distance and angles.

        :param donor:  donor atom
        :param acc:  acceptor atom
        :return:  whether this pair is a hydrogen bond
        :rtype:  bool
        """
        for donorhatom in donor.bonds:
            if not donorhatom.is_hydrogen:
                continue
            # Check the H(D)-A distance
            dist = util.distance(donorhatom.coords, acc.coords)
            if dist > DIST_CUTOFF:
                continue
            # Ensure no conflicts if H(A)s if present
            flag = True
            for acchatom in acc.bonds:
                if not acchatom.is_hydrogen:
                    continue
                flag = False
                # Check the H(D)-H(A) distance
                hdist = util.distance(donorhatom.coords, acchatom.coords)
                if hdist < 1.5:
                    continue
                # Check the H(D)-H(A)-A angle
                angle = self.get_hbond_angle(donorhatom, acchatom, acc)
                if angle < 110.0:
                    flag = True
            if not flag:
                continue
            # Check the A-D-H(D) angle
            angle = self.get_hbond_angle(acc, donor, donorhatom)
            if angle <= ANGLE_CUTOFF:
                _LOGGER.debug(f"Found HBOND! {dist:.4f} {angle:.4f}")
                return True
        # If we get here, no bond is formed
        return False

    @staticmethod
    def get_pair_energy(donor, acceptor):
        """Get the energy between two atoms

        .. todo::
           Lots of code in this function could be accelerated with
           :mod:`numpy`.

        .. todo::
           Lots of hard-coded parameters in this function that need to be
           abstracted out.

        :param donor:  the first atom in the pair
        :type donor:  Atom
        :param acceptor:  the second atom in the pair
        :type acceptor:  Atom
        :return:  the energy of the pair
        :rtype:  float
        """
        # Initialize some variables
        bump_energy = 10.0
        bump_distance = 1.5
        max_hbond_energy = -10.0
        max_ele_energy = -1.0
        adh_angle_cutoff = ANGLE_CUTOFF
        dhaha_angle_cutoff = 110.0
        max_dha_dist = DIST_CUTOFF
        max_ele_dist = 5.0
        energy = 0.0
        if not (donor.hdonor and acceptor.hacceptor):
            return energy
        # See if hydrogens are presently bonded to the acceptor and donor
        donorhs = (bond for bond in donor.bonds if bond.is_hydrogen)
        acceptorhs = [bond for bond in acceptor.bonds if bond.is_hydrogen]
        for donorhatom in donorhs:
            dist = util.distance(donorhatom.coords, acceptor.coords)
            if dist > max_dha_dist and dist < max_ele_dist:
                energy += max_ele_energy / (dist * dist)
                continue
            # Case 1: Both donor and acceptor hydrogens are present
            for acceptorhatom in acceptorhs:
                # Penalize if H(D) is too close to H(A)
                hdist = util.distance(donorhatom.coords, acceptorhatom.coords)
                if hdist < bump_distance:
                    energy += bump_energy
                    continue
                # Assign energies based on angles
                angle1 = Optimize.get_hbond_angle(acceptor, donor, donorhatom)
                if angle1 <= adh_angle_cutoff:
                    angle2 = Optimize.get_hbond_angle(
                        donorhatom, acceptorhatom, acceptor
                    )
                    if angle2 < dhaha_angle_cutoff:
                        angle2 = 1.0
                    else:
                        angle2 = (
                            dhaha_angle_cutoff - angle2
                        ) / dhaha_angle_cutoff
                    angleterm = (adh_angle_cutoff - angle1) / adh_angle_cutoff
                    energy += (
                        max_hbond_energy / pow(dist, 3) * angleterm * angle2
                    )
            # Case 2: Only donor hydrogens are present
            if not acceptorhs:
                # Assign energies based on A-D-H(D) angle alone
                angle1 = Optimize.get_hbond_angle(acceptor, donor, donorhatom)
                if angle1 <= adh_angle_cutoff:
                    angleterm = (adh_angle_cutoff - angle1) / adh_angle_cutoff
                    energy += max_hbond_energy / pow(dist, 2) * angleterm
        return energy

    def make_atom_with_no_bonds(self, atom, closeatom, addname):
        """Create an atom with no bonds.

        Called for water oxygen atoms with no current bonds.
        Uses the closeatom to place the new atom directly colinear with the
        atom and the closeatom.

        :param atom:  the oxygen atom of the water
        :type atom:  Atom
        :param closeatom: the nearby atom (donor/acceptor)
        :type closeatom:  Atom
        :param addname:  the name of the atom to add
        :type addname:  str
        """
        residue = atom.residue
        # Place along line, 1 A away
        vec = util.subtract(closeatom.coords, atom.coords)
        dist = util.distance(atom.coords, closeatom.coords)
        newcoords = [vec[i] / dist + atom.coords[i] for i in range(3)]
        residue.create_atom(addname, newcoords)
        newatom = residue.get_atom(addname)
        self.routines.cells.add_cell(newatom)
        # Set the bonds (since not in reference structure)
        if newatom not in atom.bonds:
            atom.bonds.append(newatom)
        if atom not in newatom.bonds:
            newatom.bonds.append(atom)

    @classmethod
    def make_water_with_one_bond(cls, atom, addname):
        """Add an atom to a water residue that already has one bond.

        Uses the water reference structure to align the new atom.

        .. todo::
           Does this need to be a classmethod or could it be a staticmethod?

        :param atom:  existing atom
        :type atom:  Atom
        :param addname:  name of atom to add
        :type addname:  str
        """
        residue = atom.residue
        nextatom = atom.bonds[0]
        coords = [atom.coords, nextatom.coords]
        refcoords = [
            residue.reference.map[atom.name].coords,
            residue.reference.map["H1"].coords,
        ]
        refatomcoords = residue.reference.map["H2"].coords

        # Make the atom
        newcoords = quat.find_coordinates(2, coords, refcoords, refatomcoords)
        residue.create_atom(addname, newcoords)

        # Set the bonds (since not in reference structure)
        newatom = residue.get_atom(addname)
        if newatom not in atom.bonds:
            atom.bonds.append(newatom)
        if atom not in newatom.bonds:
            newatom.bonds.append(atom)

    @classmethod
    def make_atom_with_one_bond_h(cls, atom, addname):
        """Add a hydrogen to an alcoholic donor with one existing bond.

        .. todo::
           Does this need to be a classmethod or could it be a staticmethod?

        :param atom:  existing atom
        :type atom:  Atom
        :param addname:  name of atom to add
        :type addname:  str
        """
        residue = atom.residue
        nextatom = atom.bonds[0]
        coords = [atom.coords, nextatom.coords]
        refcoords = [
            residue.reference.map[atom.name].coords,
            residue.reference.map[nextatom.name].coords,
        ]
        refatomcoords = residue.reference.map[addname].coords
        # Make the atom
        newcoords = quat.find_coordinates(2, coords, refcoords, refatomcoords)
        residue.create_atom(addname, newcoords)

    @classmethod
    def make_atom_with_one_bond_lp(cls, atom, addname):
        """Add a lone pair to an alcoholic donor with one existing bond.

        .. todo::
           Does this need to be a classmethod or could it be a staticmethod?

        :param atom:  existing atom
        :type atom:  Atom
        :param addname:  name of atom to add
        :type addname:  str
        """
        # Initialize some variables
        residue = atom.residue
        the_refname = ""
        for refname in atom.reference.bonds:
            if refname.startswith("H"):
                the_refname = refname
                break
        nextatom = atom.bonds[0]
        coords = [atom.coords, nextatom.coords]
        refcoords = [
            residue.reference.map[atom.name].coords,
            residue.reference.map[nextatom.name].coords,
        ]
        refatomcoords = residue.reference.map[the_refname].coords
        # Make the atom
        newcoords = quat.find_coordinates(2, coords, refcoords, refatomcoords)
        residue.create_atom(addname, newcoords)
        # Set the bonds (since not in reference structure)
        newatom = residue.get_atom(addname)
        if newatom not in atom.bonds:
            atom.bonds.append(newatom)
        if atom not in newatom.bonds:
            newatom.bonds.append(atom)

    def try_single_alcoholic_h(self, donor, acc, newatom):
        """Attempt to add an atom to make a hydrogen bond.

        .. todo::
           Remove some hard-coded values in this function.

        After a new bond has been added using :func:`makeAtomWithOneBond`, try
        to find the best orientation by rotating to form a hydrogen bond.
        If a bond cannot be formed, remove the :makevar:`newatom` (thereby
        returning to a single bond).

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param newatom:  new atom to add
        :type newatom:  Atom
        :return:  indication whether atom was added
        :rtype:  bool
        """
        # Initialize some variables
        besten = 999.99
        bestcoords = []
        residue = donor.residue
        pivot = donor.bonds[0]
        for _ in range(72):
            residue.rotate_tetrahedral(pivot, donor, 5.0)
            if self.is_hbond(donor, acc):
                energy = self.get_pair_energy(donor, acc)
                if energy < besten:
                    bestcoords = newatom.coords
                    besten = energy
        # If a hydrogen bond was made, set at best coordinates
        if bestcoords != []:
            newatom.x = bestcoords[0]
            newatom.y = bestcoords[1]
            newatom.z = bestcoords[2]
            self.routines.cells.add_cell(newatom)
            return True
        residue.remove_atom(newatom.name)
        return False

    def try_single_alcoholic_lp(self, acc, donor, newatom):
        """Attempt to add an atom to make a hydrogen bond.

        .. todo::
           Remove some hard-coded values in this function and figure out where
           others (e.g., "72") come from.

        After a new bond has been added using :func:`makeAtomWithOneBond`,
        ensure that a hydrogen bond has been made.
        If so, try to minimze the H(D)-A-LP angle.  If that cannot be
        minimized, ignore the bond and remove the atom.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param newatom:  new atom to add
        :type newatom:  Atom
        :return:  indication whether atom was added
        :rtype:  bool
        """
        # Initialize some variables
        residue = acc.residue
        pivot = acc.bonds[0]
        bestangle = 180.00
        bestcoords = []
        the_donorhatom = ""
        # If a hydrogen bond was made, set at best distance
        if not self.is_hbond(donor, acc):
            residue.remove_atom(newatom.name)
            return False
        # Grab the H(D) that caused the bond
        for donorhatom in donor.bonds:
            if (
                donorhatom.is_hydrogen
                and self.get_hbond_angle(acc, donor, donorhatom) < ANGLE_CUTOFF
            ):
                the_donorhatom = donorhatom
                break
        for _ in range(72):
            residue.rotate_tetrahedral(pivot, acc, 5.0)
            angle = abs(self.get_hbond_angle(the_donorhatom, acc, newatom))
            if angle < bestangle:
                bestangle = angle
                bestcoords = newatom.coords
        # Remove if geometry does not work
        if bestangle > (ANGLE_CUTOFF * 2.0):
            _LOGGER.debug(
                f"Removing due to geometry {bestangle:.2f} > "
                f"{ANGLE_CUTOFF * 2.0:.2f}"
            )
            residue.remove_atom(newatom.name)
            return False
        # Otherwise set to best coordinates
        newatom.x = bestcoords[0]
        newatom.y = bestcoords[1]
        newatom.z = bestcoords[2]
        self.routines.cells.add_cell(newatom)
        return True

    @classmethod
    def get_positions_with_two_bonds(cls, atom):
        """Return possible coordinates for new bonds.

        .. todo::
           Remove some hard-coded values in this function.

        Given a tetrahedral geometry with two existing bonds, return the two
        potential sets of coordinates that are possible for a new bond.

        :param atom:  atom to test for bonds
        :type atom:  Atom
        :return:  2-tuple of coordinates (floats)
        :rtype:  tuple
        """
        # Initialize some variables
        residue = atom.residue
        fixed = atom.bonds[0]
        rotate = atom.bonds[1]
        # Rotate by 120 degrees twice
        residue.rotate_tetrahedral(fixed, atom, 120)
        loc1 = rotate.coords
        residue.rotate_tetrahedral(fixed, atom, 120)
        loc2 = rotate.coords
        # Set rotate back to original by one more rotation
        residue.rotate_tetrahedral(fixed, atom, 120)
        return loc1, loc2

    def try_positions_with_two_bonds_h(self, donor, acc, newname, loc1, loc2):
        """Try adding a new hydrogen to the two potential locations.

        If both form hydrogen bonds, place at whatever returns the best bond as
        determined by get_pair_energy.

        :param donor:  donor atom
        :type donor:  Atom
        :param acc:  acceptor atom
        :type acc:  Atom
        :param newname:  new atom name
        :type newname:  str
        :param loc1:  first coordinate to try
        :type loc1:  (float, float, float)
        :param loc2:  first coordinate to try
        :type loc2:  (float, float, float)
        :return:  indication whether atom was added
        :rtype:  bool
        """
        # Initialize some variables
        besten = 999.99
        bestcoords = []
        residue = donor.residue
        # Try the first position
        residue.create_atom(newname, loc1)
        if self.is_hbond(donor, acc):
            besten = self.get_pair_energy(donor, acc)
            bestcoords = loc1
        # Try the second
        newatom = residue.get_atom(newname)
        newatom.x = loc2[0]
        newatom.y = loc2[1]
        newatom.z = loc2[2]
        if self.is_hbond(donor, acc):
            energy = self.get_pair_energy(donor, acc)
            if energy < besten:
                bestcoords = loc2
        # Set at best coords
        if bestcoords != []:
            newatom.x = bestcoords[0]
            newatom.y = bestcoords[1]
            newatom.z = bestcoords[2]
            self.routines.cells.add_cell(newatom)
            return 1
        residue.remove_atom(newname)
        return 0

    def try_positions_with_two_bonds_lp(self, acc, donor, newname, loc1, loc2):
        """Attempt to place a :abbr:`LP (lone pair)` on an atom.

        Try placing an LP on a tetrahedral geometry with two existing bonds.
        If this isn't a hydrogen bond it can return - otherwise ensure that the
        H(D)-A-LP angle is minimized.

        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param newname:  name for lone pair
        :type newname:  str
        :param loc1:  first location to try
        :type loc1:  (float, float, float)
        :param loc2:  second location to try
        :type loc2:  (float, float, float)
        :return:  indication of whether LP was added
        :rtype:  bool
        """
        # Initialize some variables
        bestangle = 180.00
        bestcoords = []
        residue = acc.residue
        the_donorhatom = ""
        # If the donor/acceptor pair is not an hbond return
        if not self.is_hbond(donor, acc):
            return False
        # Grab the H(D) that caused the bond
        for donorhatom in donor.bonds:
            if (
                donorhatom.is_hydrogen
                and self.get_hbond_angle(acc, donor, donorhatom) < ANGLE_CUTOFF
            ):
                the_donorhatom = donorhatom
                break
        # Try the first position
        residue.create_atom(newname, loc1)
        newatom = residue.get_atom(newname)
        angle = abs(self.get_hbond_angle(the_donorhatom, acc, newatom))
        if angle < bestangle:
            bestangle = angle
            bestcoords = loc1
        # Try the second
        newatom.x = loc2[0]
        newatom.y = loc2[1]
        newatom.z = loc2[2]
        angle = self.get_hbond_angle(the_donorhatom, acc, newatom)
        if angle < bestangle:
            bestcoords = loc2
        # Remove if geometry does not work
        if bestangle > (ANGLE_CUTOFF * 2.0):
            residue.remove_atom(newname)
            return False
        # Otherwise set at best coords
        newatom.x = bestcoords[0]
        newatom.y = bestcoords[1]
        newatom.z = bestcoords[2]
        self.routines.cells.add_cell(newatom)
        # Set the bonds (since not in reference structure)
        if newatom not in acc.bonds:
            acc.bonds.append(newatom)
        if acc not in newatom.bonds:
            newatom.bonds.append(acc)
        return True

    @classmethod
    def get_position_with_three_bonds(cls, atom):
        """Find position for last bond in tetrahedral geometry.

        .. todo::
           Should this be a staticmethod rather than a classmethod?

        .. todo::
           Remove hard-coded values in function.

        If there's three bonds in a tetrahedral geometry, there's only one
        available position.
        Find that position.

        :param atom:  atom to check
        :type atom:  Atom
        :return:  coordinates
        :rtype:  (float, float, float)
        """
        # Initialize some variables
        residue = atom.residue
        pivot = atom.bonds[0]
        rot1 = atom.bonds[1]
        rot2 = atom.bonds[2]
        # Find the two new positions
        residue.rotate_tetrahedral(pivot, atom, 120)
        newcoords1 = rot1.coords
        residue.rotate_tetrahedral(pivot, atom, 120)
        newcoords2 = rot1.coords
        residue.rotate_tetrahedral(pivot, atom, 120)
        # Determine which is unoccupied
        if util.distance(rot2.coords, newcoords1) > 0.1:
            return newcoords1
        return newcoords2

    def try_positions_three_bonds_h(self, donor, acc, newname, loc):
        """Try making a hydrogen bond with the lone available position.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param newname:  name for new atom
        :type newname:  str
        :return:  indication of whether atom was added
        :rtype:  bool
        """
        residue = donor.residue
        residue.create_atom(newname, loc)
        if self.is_hbond(donor, acc):
            newatom = residue.get_atom(newname)
            self.routines.cells.add_cell(newatom)
            return True
        residue.remove_atom(newname)
        return False

    def try_positions_three_bonds_lp(self, acc, donor, newname, loc):
        """Make a hydrogen bond in the only position possible.

        Try making a hydrogen bond using the lone available hydrogen position.

        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param newname:  new atom name
        :type newname:  str
        :param loc:  location for atom (coordinates)
        :type loc:  (float, float, float)
        :return:  indication whether atom was added
        :rtype:  bool
        """
        residue = acc.residue
        if not self.is_hbond(donor, acc):
            return False
        # Grab the H(D) that caused the bond
        for donorhatom in donor.bonds:
            if (
                donorhatom.is_hydrogen
                and self.get_hbond_angle(acc, donor, donorhatom) < ANGLE_CUTOFF
            ):
                the_donorhatom = donorhatom
                break
        residue.create_atom(newname, loc)
        newatom = residue.get_atom(newname)
        # Remove if geometry does not work
        # WARNING: Nathan? What if the the_donorhatom is not set?
        angle = abs(self.get_hbond_angle(the_donorhatom, acc, newatom))
        if angle > (ANGLE_CUTOFF * 2.0):
            residue.remove_atom(newname)
            return False
        # Otherwise keep it
        newatom = residue.get_atom(newname)
        self.routines.cells.add_cell(newatom)
        # Set the bonds (since not in reference structure)
        if newatom not in acc.bonds:
            acc.bonds.append(newatom)
        if acc not in newatom.bonds:
            newatom.bonds.append(acc)
        return True


class OptimizationHolder:
    """A holder class for the XML parser."""

    def __init__(self):
        self.name = ""
        self.map = {}
        self.opttype = ""
        self.optangle = ""

    def __str__(self):
        text = f"{self.name}\n"
        text += f"Type: {self.opttype}\n"
        if self.optangle != "":
            text += f"Optimization Angle: {self.optangle}\n"
        text += "Atoms: \n"
        for atomname in self.map:
            text += f"\t{self.map[atomname]}\n"
        return text
